import torch
import torch.nn.functional as F

__all__ = [
    'pairwise_cosine_similarity',
    'matching_log_probas',
    'matching_probas',
    'matching_loss'
]


def pairwise_cosine_similarity(embeddings1, embeddings2, eps=1e-8):
    r"""Computes the pairwise cosine similarity between two tensors of embeddings.

    Parameters
    ----------
    embeddings1 : `torch.Tensor` instance
        A tensor containing embeddings with shape
        `(batch_size, N, embedding_size)`.

    embeddings2 : `torch.Tensor` instance
        A tensor containing embeddings with shape
        `(batch_size, M, embedding_size)`.

    eps: float (default: 1e-8)
        Small value to avoid division by zero.

    Returns
    -------
    similarities : `torch.Tensor` instance
        A tensor containing the pairwise cosine similarities between the vectors
        in `embeddings1` and `embeddings2`. This tensor has shape
        `(batch_size, N, M)`.

    Notes
    -----
    The cosine similarity is computed as

        .. math ::
            \text{similarity} = \dfrac{x_1 \cdot x_2}{\max(\Vert x_1 \Vert _2 \cdot \Vert x_2 \Vert _2, \epsilon)}
    """
    sq_norm1 = torch.sum(embeddings1 ** 2, dim=2, keepdim=True)
    sq_norm2 = torch.sum(embeddings2 ** 2, dim=2).unsqueeze(1)
    dot_product = torch.bmm(embeddings1, embeddings2.transpose(1, 2))
    inverse_norm = torch.rsqrt(torch.clamp(sq_norm1 * sq_norm2, min=eps ** 2))
    return dot_product * inverse_norm


def matching_log_probas(embeddings, targets, test_embeddings, num_classes, eps=1e-8):
    """Computes the log-probability of test samples given the training dataset
    for the matching network [1].

    Parameters
    ----------
    embeddings : `torch.Tensor` instance
        A tensor containing the embeddings of the train/support inputs. This
        tensor has shape `(batch_size, num_train_samples, embedding_size)`.

    targets : `torch.LongTensor` instance
        A tensor containing the targets of the train/support dataset. This tensor
        has shape `(batch_size, num_train_samples)`.

    test_embeddings : `torch.Tensor` instance
        A tensor containing the embeddings of the test/query inputs. This tensor
        has shape `(batch_size, num_test_samples, embedding_size)`.

    num_classes : int
        Number of classes (i.e. `N` in "N-way classification") in the
        classification task.

    eps : float (default: 1e-8)
        Small value to avoid division by zero.

    Returns
    -------
    log_probas : `torch.Tensor` instance
        A tensor containing the log-probabilities of the test samples given the
        training dataset for the matching network. This tensor has shape
        `(batch_size, num_classes, num_test_samples)`.

    References
    ----------
    .. [1] Vinyals, O., Blundell, C., Lillicrap, T. and Wierstra, D. (2016).
           Matching Networks for One Shot Learning. In Advances in Neural
           Information Processing Systems (pp. 3630-3638) (https://arxiv.org/abs/1606.04080)
    """
    batch_size, num_samples, _ = test_embeddings.shape
    similarities = pairwise_cosine_similarity(embeddings, test_embeddings, eps=eps)
    logsumexp = torch.logsumexp(similarities, dim=1, keepdim=True)

    max_similarities, _ = torch.max(similarities, dim=1, keepdim=True)
    exp_similarities = torch.exp(similarities - max_similarities)

    sum_exp = exp_similarities.new_zeros((batch_size, num_classes, num_samples))
    indices = targets.unsqueeze(-1).expand_as(exp_similarities)
    sum_exp.scatter_add_(1, indices, exp_similarities)

    return torch.log(sum_exp) + max_similarities - logsumexp


def matching_probas(embeddings, targets, test_embeddings, num_classes, eps=1e-8):
    """Computes the probability of test samples given the training dataset for
    the matching network [1].

    Parameters
    ----------
    embeddings : `torch.Tensor` instance
        A tensor containing the embeddings of the train/support inputs. This
        tensor has shape `(batch_size, num_train_samples, embedding_size)`.

    targets : `torch.LongTensor` instance
        A tensor containing the targets of the train/support dataset. This tensor
        has shape `(batch_size, num_train_samples)`.

    test_embeddings : `torch.Tensor` instance
        A tensor containing the embeddings of the test/query inputs. This tensor
        has shape `(batch_size, num_test_samples, embedding_size)`.

    num_classes : int
        Number of classes (i.e. `N` in "N-way classification") in the
        classification task.

    eps : float (default: 1e-8)
        Small value to avoid division by zero.

    Returns
    -------
    probas : `torch.Tensor` instance
        A tensor containing the probabilities of the test samples given the
        training dataset for the matching network. This tensor has shape
        `(batch_size, num_classes, num_test_samples)`.

    References
    ----------
    .. [1] Vinyals, O., Blundell, C., Lillicrap, T. and Wierstra, D. (2016).
           Matching Networks for One Shot Learning. In Advances in Neural
           Information Processing Systems (pp. 3630-3638) (https://arxiv.org/abs/1606.04080)
    """
    log_probas = matching_log_probas(embeddings,
                                     targets,
                                     test_embeddings,
                                     num_classes,
                                     eps=eps)
    return log_probas.exp()


def matching_loss(train_embeddings,
                  train_targets,
                  test_embeddings,
                  test_targets,
                  num_classes,
                  eps=1e-8,
                  **kwargs):
    """Compute the loss (i.e. negative log-likelihood) for the matching network
    on the test/query samples [1].

    Parameters
    ----------
    train_embeddings : `torch.Tensor` instance
        A tensor containing the embeddings of the train/support inputs. This
        tensor has shape `(batch_size, num_train_samples, embedding_size)`.

    train_targets : `torch.LongTensor` instance
        A tensor containing the targets of the train/support dataset. This tensor
        has shape `(batch_size, num_train_samples)`.

    test_embeddings : `torch.Tensor` instance
        A tensor containing the embeddings of the test/query inputs. This tensor
        has shape `(batch_size, num_test_samples, embedding_size)`.

    test_targets : `torch.LongTensor` instance
        A tensor containing the targets of the test/query dataset. This tensor
        has shape `(batch_size, num_test_samples)`.

    num_classes : int
        Number of classes (i.e. `N` in "N-way classification") in the
        classification task.

    eps : float (default: 1e-8)
        Small value to avoid division by zero.

    kwargs :
        Additional keyword arguments to be forwarded to the loss function. See
        `torch.nn.functional.cross_entropy` for details.

    Returns
    -------
    loss : `torch.Tensor` instance
        A tensor containing the loss for the matching network.

    References
    ----------
    .. [1] Vinyals, O., Blundell, C., Lillicrap, T. and Wierstra, D. (2016).
           Matching Networks for One Shot Learning. In Advances in Neural
           Information Processing Systems (pp. 3630-3638) (https://arxiv.org/abs/1606.04080)
    """
    logits = matching_log_probas(train_embeddings,
                                 train_targets,
                                 test_embeddings,
                                 num_classes,
                                 eps=eps)
    return F.nll_loss(logits, test_targets, **kwargs)



import numpy as np

from torchmeta.utils.data import Task, MetaDataset
from torchmeta.toy.sinusoid import SinusoidTask


class SinusoidAndLine(MetaDataset):
    """
    Simple multimodal regression task, based on sinusoids and lines, as
    introduced in [1].

    Parameters
    ----------
    num_samples_per_task : int
        Number of examples per task.

    num_tasks : int (default: 1,000,000)
        Overall number of tasks to sample.

    noise_std : float, optional
        Amount of noise to include in the targets for each task. If `None`, then
        nos noise is included, and the target is either a sine function, or a
        linear function of the input.

    transform : callable, optional
        A function/transform that takes a numpy array of size (1,) and returns a
        transformed version of the input.

    target_transform : callable, optional
        A function/transform that takes a numpy array of size (1,) and returns a
        transformed version of the target.

    dataset_transform : callable, optional
        A function/transform that takes a dataset (ie. a task), and returns a 
        transformed version of it. E.g. `torchmeta.transforms.ClassSplitter()`.

    Notes
    -----
    The tasks are created randomly as either random sinusoid functions, or
    random linear functions. The amplitude of the sinusoids varies within
    [0.1, 5.0] and the phase within [0, pi]. The slope and intercept of the lines
    vary in [-3.0, 3.0]. The inputs are sampled uniformly in [-5.0, 5.0]. Due to
    the way PyTorch handles datasets, the number of tasks to be sampled needs to
    be fixed ahead of time (with `num_tasks`). This will typically be equal to
    `meta_batch_size * num_batches`.

    References
    ----------
    .. [1] Finn C., Xu K., Levine S. (2018). Probabilistic Model-Agnostic
           Meta-Learning. In Advances in Neural Information Processing Systems
           (https://arxiv.org/abs/1806.02817)
    """
    def __init__(self, num_samples_per_task, num_tasks=1000000,
                 noise_std=None, transform=None, target_transform=None,
                 dataset_transform=None):
        super(SinusoidAndLine, self).__init__(meta_split='train',
            target_transform=target_transform, dataset_transform=dataset_transform)
        self.num_samples_per_task = num_samples_per_task
        self.num_tasks = num_tasks
        self.noise_std = noise_std
        self.transform = transform

        self._input_range = np.array([-5.0, 5.0])
        self._amplitude_range = np.array([0.1, 5.0])
        self._phase_range = np.array([0, np.pi])
        self._slope_range = np.array([-3.0, 3.0])
        self._intercept_range = np.array([-3.0, 3.0])

        
        self._is_sinusoid = None
        self._amplitudes = None
        self._phases = None
        self._slopes = None
        self._intercepts = None

    @property
    def amplitudes(self):
        if self._amplitudes is None:
            self._amplitudes = self.np_random.uniform(self._amplitude_range[0],
                self._amplitude_range[1], size=self.num_tasks)
        return self._amplitudes

    @property
    def phases(self):
        if self._phases is None:
            self._phases = self.np_random.uniform(self._phase_range[0],
                self._phase_range[1], size=self.num_tasks)
        return self._phases

    @property
    def slopes(self):
        if self._slopes is None:
            self._slopes = self.np_random.uniform(self._slope_range[0],
                self._slope_range[1], size=self.num_tasks)
        return self._slopes

    @property
    def intercepts(self):
        if self._intercepts is None:
            self._intercepts = self.np_random.uniform(self._intercept_range[0],
                self._intercept_range[1], size=self.num_tasks)
        return self._intercepts

    @property
    def is_sinusoid(self):
        if self._is_sinusoid is None:
            self._is_sinusoid = np.zeros((self.num_tasks,), dtype=np.bool_)
            self._is_sinusoid[self.num_tasks // 2:] = True
            self.np_random.shuffle(self._is_sinusoid)
        return self._is_sinusoid

    def __len__(self):
        return self.num_tasks

    def __getitem__(self, index):
        if self.is_sinusoid[index]:
            amplitude, phase = self.amplitudes[index], self.phases[index]
            task = SinusoidTask(index, amplitude, phase, self._input_range,
                self.noise_std, self.num_samples_per_task, self.transform,
                self.target_transform, np_random=self.np_random)
        else:
            slope, intercept = self.slopes[index], self.intercepts[index]
            task = LinearTask(index, slope, intercept, self._input_range,
                self.noise_std, self.num_samples_per_task, self.transform,
                self.target_transform, np_random=self.np_random)

        if self.dataset_transform is not None:
            task = self.dataset_transform(task)

        return task


class LinearTask(Task):
    def __init__(self, index, slope, intercept, input_range, noise_std,
                 num_samples, transform=None, target_transform=None,
                 np_random=None):
        super(LinearTask, self).__init__(index, None) # Regression task
        self.slope = slope
        self.intercept = intercept
        self.input_range = input_range
        self.num_samples = num_samples
        self.noise_std = noise_std

        self.transform = transform
        self.target_transform = target_transform

        if np_random is None:
            np_random = np.random.RandomState(None)

        self._inputs = np_random.uniform(input_range[0], input_range[1],
            size=(num_samples, 1))
        self._targets = intercept + slope * self._inputs
        if (noise_std is not None) and (noise_std > 0.):
            self._targets += noise_std * np_random.randn(num_samples, 1)

    def __len__(self):
        return self.num_samples

    def __getitem__(self, index):
        input, target = self._inputs[index], self._targets[index]

        if self.transform is not None:
            input = self.transform(input)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return (input, target)



import numpy as np
from PIL import Image
import os
import io
import json
import glob
import h5py

from torchmeta.utils.data import Dataset, ClassDataset, CombinationMetaDataset
# QKFIX: See torchmeta.datasets.utils for more informations
from torchmeta.datasets.utils import download_file_from_google_drive
from torchmeta.datasets.utils import get_asset


class DoubleMNIST(CombinationMetaDataset):
    """
    The Double MNIST dataset, introduced in [1]. This dataset is based on
    the MNIST dataset [2]. It consists of sampled images from MNIST
    that are put together to create images with multiple digits. It contains
    100,000 images from 100 different classes (1000 images per class, for the 
    numbers 00 to 99).

    Parameters
    ----------
    root : string
        Root directory where the dataset folder `doublemnist` exists.

    num_classes_per_task : int
        Number of classes per tasks. This corresponds to "N" in "N-way" 
        classification.

    meta_train : bool (default: `False`)
        Use the meta-train split of the dataset. If set to `True`, then the
        arguments `meta_val` and `meta_test` must be set to `False`. Exactly one 
        of these three arguments must be set to `True`.

    meta_val : bool (default: `False`)
        Use the meta-validation split of the dataset. If set to `True`, then the 
        arguments `meta_train` and `meta_test` must be set to `False`. Exactly
        one of these three arguments must be set to `True`.

    meta_test : bool (default: `False`)
        Use the meta-test split of the dataset. If set to `True`, then the 
        arguments `meta_train` and `meta_val` must be set to `False`. Exactly
        one of these three arguments must be set to `True`.

    meta_split : string in {'train', 'val', 'test'}, optional
        Name of the split to use. This overrides the arguments `meta_train`, 
        `meta_val` and `meta_test` if all three are set to `False`.

    transform : callable, optional
        A function/transform that takes a `PIL` image, and returns a transformed 
        version. See also `torchvision.transforms`.

    target_transform : callable, optional
        A function/transform that takes a target, and returns a transformed 
        version. See also `torchvision.transforms`.

    dataset_transform : callable, optional
        A function/transform that takes a dataset (ie. a task), and returns a 
        transformed version of it. E.g. `torchmeta.transforms.ClassSplitter()`.

    class_augmentations : list of callable, optional
        A list of functions that augment the dataset with new classes. These
        classes are transformations of existing classes. E.g.
        `torchmeta.transforms.HorizontalFlip()`.

    download : bool (default: `False`)
        If `True`, downloads the pickle files and processes the dataset in the
        root directory (under the `doublemnist` folder). If the dataset is
        already available, this does not download/process the dataset again.

    Notes
    -----
    The dataset is downloaded from the Multi-digit MNIST repository
    [1](https://github.com/shaohua0116/MultiDigitMNIST). The dataset contains
    images (MNIST double digits) from 100 classes, for the numbers 00 to 99.
    The meta train/validation/test splits are 64/16/20 classes.
    The splits are taken from [1].

    References
    ----------
    .. [1] Sun, S. (2019). Multi-digit MNIST for Few-shot Learning.
    (https://github.com/shaohua0116/MultiDigitMNIST)

    .. [2] LeCun, Y., Cortes, C., and Burges, CJ. (2010). MNIST Handwritten
    Digit Database. (http://yann.lecun.com/exdb/mnist)

    """
    def __init__(self, root, num_classes_per_task=None, meta_train=False,
                 meta_val=False, meta_test=False, meta_split=None,
                 transform=None, target_transform=None, dataset_transform=None,
                 class_augmentations=None, download=False):
        dataset = DoubleMNISTClassDataset(root,
            meta_train=meta_train, meta_val=meta_val,
            meta_test=meta_test, meta_split=meta_split, transform=transform,
            class_augmentations=class_augmentations, download=download)
        super(DoubleMNIST, self).__init__(dataset, num_classes_per_task,
            target_transform=target_transform,
            dataset_transform=dataset_transform)


class DoubleMNISTClassDataset(ClassDataset):
    folder = 'doublemnist'
    # Google Drive ID from https://github.com/shaohua0116/MultiDigitMNIST
    gdrive_id = '1MqQCdLt9TVE3joAMw4FwJp_B8F-htrAo'
    zip_filename = 'double_mnist_seed_123_image_size_64_64.zip'
    zip_md5 = '6d8b185c0cde155eb39d0e3615ab4f23'

    filename = '{0}_data.hdf5'
    filename_labels = '{0}_labels.json'

    image_folder = 'double_mnist_seed_123_image_size_64_64'

    def __init__(self, root, meta_train=False, meta_val=False, meta_test=False,
                 meta_split=None, transform=None, class_augmentations=None,
                 download=False):
        super(DoubleMNISTClassDataset, self).__init__(meta_train=meta_train,
            meta_val=meta_val, meta_test=meta_test, meta_split=meta_split,
            class_augmentations=class_augmentations)

        self.root = os.path.join(os.path.expanduser(root), self.folder)
        self.transform = transform

        self.split_filename = os.path.join(self.root,
            self.filename.format(self.meta_split))
        self.split_filename_labels = os.path.join(self.root,
            self.filename_labels.format(self.meta_split))

        self._data_file = None
        self._data = None
        self._labels = None

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError('Double MNIST integrity check failed')
        self._num_classes = len(self.labels)

    def __getitem__(self, index):
        label = self.labels[index % self.num_classes]
        data = self.data[label]
        transform = self.get_transform(index, self.transform)
        target_transform = self.get_target_transform(index)

        return DoubleMNISTDataset(index, data, label, transform=transform,
                                  target_transform=target_transform)

    @property
    def num_classes(self):
        return self._num_classes

    @property
    def data(self):
        if self._data is None:
            self._data_file = h5py.File(self.split_filename, 'r')
            self._data = self._data_file['datasets']
        return self._data
    
    @property
    def labels(self):
        if self._labels is None:
            with open(self.split_filename_labels, 'r') as f:
                self._labels = json.load(f)
        return self._labels

    def _check_integrity(self):
        return (os.path.isfile(self.split_filename)
            and os.path.isfile(self.split_filename_labels))

    def close(self):
        if self._data_file is not None:
            self._data_file.close()
            self._data_file = None
            self._data = None

    def download(self):
        import zipfile
        import shutil
        import glob
        from tqdm import tqdm

        if self._check_integrity():
            return

        zip_filename = os.path.join(self.root, self.zip_filename)
        if not os.path.isfile(zip_filename):
            download_file_from_google_drive(self.gdrive_id, self.root,
                self.zip_filename, md5=self.zip_md5)

        zip_foldername = os.path.join(self.root, self.image_folder)
        if not os.path.isdir(zip_foldername):
            with zipfile.ZipFile(zip_filename, 'r') as f:
                for member in tqdm(f.infolist(), desc='Extracting '):
                    try:
                        f.extract(member, self.root)
                    except zipfile.BadZipFile:
                        print('Error: Zip file is corrupted')

        for split in ['train', 'val', 'test']:
            filename = os.path.join(self.root, self.filename.format(split))
            if os.path.isfile(filename):
                continue

            labels = get_asset(self.folder, '{0}.json'.format(split))
            labels_filename = os.path.join(self.root,
                                           self.filename_labels.format(split))
            with open(labels_filename, 'w') as f:
                json.dump(labels, f)

            image_folder = os.path.join(zip_foldername, split)

            with h5py.File(filename, 'w') as f:
                group = f.create_group('datasets')
                dtype = h5py.special_dtype(vlen=np.uint8)
                for i, label in enumerate(tqdm(labels, desc=filename)):
                    images = glob.glob(os.path.join(image_folder, label,
                                                    '*.png'))
                    images.sort()
                    dataset = group.create_dataset(label, (len(images),),
                                                   dtype=dtype)
                    for i, image in enumerate(images):
                        with open(image, 'rb') as f:
                            array = bytearray(f.read())
                            dataset[i] = np.asarray(array, dtype=np.uint8)

        if os.path.isdir(zip_foldername):
            shutil.rmtree(zip_foldername)


class DoubleMNISTDataset(Dataset):
    def __init__(self, index, data, label,
                 transform=None, target_transform=None):
        super(DoubleMNISTDataset, self).__init__(index, transform=transform,
                                                 target_transform=target_transform)
        self.data = data
        self.label = label

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        image = Image.open(io.BytesIO(self.data[index])).convert('RGB')
        target = self.label

        if self.transform is not None:
            image = self.transform(image)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return (image, target)



import pytest

import os
import torch
from torch.utils.data import Dataset, DataLoader

from torchmeta.datasets import helpers
from torchmeta.utils.data import NonEpisodicWrapper

is_local = (os.getenv('TORCHMETA_DATA_FOLDER') is not None)


@pytest.mark.skipif(not is_local, reason='Requires datasets downloaded locally')
@pytest.mark.parametrize('name', helpers.__all__)
@pytest.mark.parametrize('split', ['train', 'val', 'test'])
def test_datasets_helpers_wrapper(name, split):
    function = getattr(helpers, name)
    folder = os.getenv('TORCHMETA_DATA_FOLDER')
    download = bool(os.getenv('TORCHMETA_DOWNLOAD', False))

    dataset = function(folder,
                       ways=5,
                       shots=1,
                       test_shots=15,
                       meta_split=split,
                       download=download)

    wrapped_dataset = NonEpisodicWrapper(dataset)

    assert isinstance(wrapped_dataset, Dataset)

    img, label = wrapped_dataset[0]

    dataloader = DataLoader(wrapped_dataset, batch_size=4)
    inputs, targets = next(iter(dataloader))

    assert isinstance(inputs, torch.Tensor)
    assert inputs.ndim == 4
    assert inputs.shape[0] == 4

    assert len(targets) == 2
    assert isinstance(targets[1], torch.Tensor)



import torch
import torch.nn.functional as F

from torchmeta.utils.prototype import get_prototypes

__all__ = ['hardness_metric']


def _pad_images(inputs, size=(224, 224), **kwargs):
    height, width = inputs.shape[-2:]
    pad_height, pad_width = (size[0] - height) // 2, (size[1] - width) // 2
    padding = (pad_width, size[1] - width - pad_width,
               pad_height, size[0] - height - pad_height)
    return F.pad(inputs, padding, **kwargs)


def hardness_metric(batch, num_classes):
    """Hardness metric of an episode, as defined in [1].

    Parameters
    ----------
    batch : dict
        The batch of tasks over which the metric is computed. The batch of tasks
        is a dictionary containing the keys `train` (or `support`) and `test`
        (or `query`). This is typically the output of `BatchMetaDataLoader`.

    num_classes : int
        The number of classes in the classification task. This corresponds to
        the number of ways in an `N`-way classification problem.

    Returns
    -------
    metric : `torch.FloatTensor` instance
        Values of the hardness metric for each task in the batch.

    References
    ----------
    .. [1] Dhillon, G. S., Chaudhari, P., Ravichandran, A. and Soatto S. (2019).
           A Baseline for Few-Shot Image Classification. (https://arxiv.org/abs/1909.02729)
    """
    if ('train' not in batch) and ('support' not in batch):
        raise ValueError('The tasks do not contain any training/support set. '
                         'Make sure the tasks contain either the "train" or the '
                         '"support" key.')
    if ('test' not in batch) and ('query' not in batch):
        raise ValueError('The tasks do not contain any test/query set. Make '
                         'sure the tasks contain either the "test" of the '
                         '"query" key.')

    train = 'train' if ('train' in batch) else 'support'
    test = 'test' if ('test' in batch) else 'query'

    with torch.no_grad():
        # Load a pre-trained backbone Resnet-152 model from PyTorch Hub
        backbone = torch.hub.load('pytorch/vision:v0.5.0',
                                  'resnet152',
                                  pretrained=True,
                                  verbose=False)
        backbone.eval()

        train_inputs, train_targets = batch[train]
        test_inputs, test_targets = batch[test]
        batch_size, num_images, num_channels = train_inputs.shape[:3]
        num_test_images = test_inputs.size(1)

        backbone.to(device=train_inputs.device)

        if num_channels != 3:
            raise ValueError('The images must be RGB images.')

        # Pad the images so that they are compatible with the pre-trained model
        padded_train_inputs = _pad_images(train_inputs,
            size=(224, 224), mode='constant', value=0.)
        padded_test_inputs = _pad_images(test_inputs,
            size=(224, 224), mode='constant', value=0.)

        # Compute the features from the logits returned by the pre-trained
        # model on the train/support examples. These features are z(x, theta)_+,
        # averaged for each class
        train_logits = backbone(padded_train_inputs.view(-1, 3, 224, 224))
        train_logits = F.relu(train_logits.view(batch_size, num_images, -1))
        train_features = get_prototypes(train_logits, train_targets, num_classes)

        # Get the weights by normalizing the features
        weights = F.normalize(train_features, p=2, dim=2)

        # Compute and normalize the logits of the test/query examples
        test_logits = backbone(padded_test_inputs.view(-1, 3, 224, 224))
        test_logits = test_logits.view(batch_size, num_test_images, -1)
        test_logits = F.normalize(test_logits, p=2, dim=2)

        # Compute the log probabilities of the test/query examples
        test_logits = torch.bmm(weights, test_logits.transpose(1, 2))
        test_log_probas = -F.cross_entropy(test_logits, test_targets,
                                           reduction='none')

        # Compute the log-odds ratios for each image of the test/query set
        log_odds_ratios = torch.log1p(-test_log_probas.exp()) - test_log_probas

    return torch.mean(log_odds_ratios, dim=1)



import torchvision.transforms.functional as F

class Rotation(object):
    def __init__(self, angle, resample=False, expand=False, center=None):
        super(Rotation, self).__init__()
        if isinstance(angle, (list, tuple)):
            self._angles = angle
            self.angle = None
        else:
            self._angles = [angle]
            self.angle = angle
            if angle % 360 == 0:
                import warnings
                warnings.warn('Applying a rotation of {0} degrees (`{1}`) as a '
                    'class augmentation on a dataset is equivalent to the original '
                    'dataset.'.format(angle, self), UserWarning, stacklevel=2)

        self.resample = resample
        self.expand = expand
        self.center = center

    def __iter__(self):
        return iter(Rotation(angle, resample=self.resample, expand=self.expand,
            center=self.center) for angle in self._angles)

    def __call__(self, image):
        if self.angle is None:
            raise ValueError('The value of the angle is unspecified.')
        # QKFIX: Explicitly compute the pixel fill value due to an
        # incompatibility between Torchvision 0.5 and Pillow 7.0.0
        # https://github.com/pytorch/vision/issues/1759#issuecomment-583826810
        # Will be fixed in Torchvision 0.6
        fill = tuple([0] * len(image.getbands()))
        return F.rotate(image, self.angle % 360, self.resample,
                        self.expand, self.center, fill=fill)

    def __hash__(self):
        return hash(repr(self))

    def __eq__(self, other):
        if (self.angle is None) or (other.angle is None):
            return self._angles == other._angles
        return (self.angle % 360) == (other.angle % 360)

    def __repr__(self):
        if self.angle is None:
            return 'Rotation({0})'.format(', '.join(map(str, self._angles)))
        else:
            return 'Rotation({0})'.format(self.angle % 360)

    def __str__(self):
        if self.angle is None:
            return 'Rotation({0})'.format(', '.join(map(str, self._angles)))
        else:
            return 'Rotation({0})'.format(self.angle)

class HorizontalFlip(object):
    def __iter__(self):
        return iter([HorizontalFlip()])

    def __call__(self, image):
        return F.hflip(image)

    def __repr__(self):
        return 'HorizontalFlip()'

class VerticalFlip(object):
    def __iter__(self):
        return iter([VerticalFlip()])

    def __call__(self, image):
        return F.vflip(image)

    def __repr__(self):
        return 'VerticalFlip()'



import pytest

import torch
import torch.nn as nn

from torchmeta.modules import MetaLinear, MetaSequential
from torchmeta.modules import DataParallel
from torchmeta.utils.gradient_based import gradient_update_parameters

is_multi_gpu = (torch.cuda.device_count() > 1)

def model():
    model = MetaSequential(
        MetaLinear(2, 3, bias=True),
        nn.ReLU(),
        MetaLinear(3, 1, bias=False))

    return model

def linear_model():
    return MetaLinear(2, 1)

def params(prefix=''):
    weight_0 = torch.tensor([
        [0.02, 0.03],
        [0.05, 0.07],
        [0.11, 0.13]], dtype=torch.float32)
    bias_0 = torch.tensor([0.17, 0.19, 0.23], dtype=torch.float32)
    weight_2 = torch.tensor([[0.29, 0.31, 0.37]], dtype=torch.float32)

    return ({
        f'{prefix}0.weight': weight_0,
        f'{prefix}0.bias': bias_0,
        f'{prefix}2.weight': weight_2
    })

def linear_params(prefix=''):
    weight = torch.tensor([[0.02, 0.03]], dtype=torch.float32)
    bias = torch.tensor([0.05], dtype=torch.float32)

    return {f'{prefix}weight': weight, f'{prefix}bias': bias}

def _dict_to_device(params, device):
    return dict((name, param.to(device=device))
        for (name, param) in params.items())


@pytest.mark.skipif(not is_multi_gpu, reason='Requires Multi-GPU support')
@pytest.mark.parametrize('model', [linear_model(), model()])
def test_dataparallel(model):
    device = torch.device('cuda:0')
    model = DataParallel(model)
    model.to(device=device)

    inputs = torch.rand(5, 2).to(device=device)
    outputs = model(inputs)

    assert outputs.shape == (5, 1)
    assert outputs.device == device


@pytest.mark.skipif(not is_multi_gpu, reason='Requires Multi-GPU support')
@pytest.mark.parametrize('model,params', [
    (linear_model(), linear_params('')),
    (linear_model(), linear_params('module.')),
    (model(), params('')),
    (model(), params('module.'))
])
def test_dataparallel_params(model, params):
    device = torch.device('cuda:0')
    model = DataParallel(model)
    model.to(device=device)
    params = _dict_to_device(params, device)

    inputs = torch.rand(5, 2).to(device=device)
    outputs = model(inputs, params=params)

    assert outputs.shape == (5, 1)
    assert outputs.device == device


@pytest.mark.skipif(not is_multi_gpu, reason='Requires Multi-GPU support')
@pytest.mark.parametrize('model', [linear_model(), model()])
def test_dataparallel_params_none(model):
    device = torch.device('cuda:0')
    model = DataParallel(model)
    model.to(device=device)

    inputs = torch.rand(5, 2).to(device=device)
    outputs = model(inputs, params=None)

    assert outputs.shape == (5, 1)
    assert outputs.device == device


@pytest.mark.skipif(not is_multi_gpu, reason='Requires Multi-GPU support')
@pytest.mark.parametrize('model', [linear_model(), model()])
def test_dataparallel_params_maml(model):
    device = torch.device('cuda:0')
    model = DataParallel(model)
    model.to(device=device)

    train_inputs = torch.rand(5, 2).to(device=device)
    train_outputs = model(train_inputs)

    inner_loss = train_outputs.sum()  # Dummy loss
    params = gradient_update_parameters(model, inner_loss)

    test_inputs = torch.rand(5, 2).to(device=device)
    test_outputs = model(test_inputs, params=params)

    assert test_outputs.shape == (5, 1)
    assert test_outputs.device == device

    outer_loss = test_outputs.sum()  # Dummy loss
    outer_loss.backward()



"""
;==========================================
; Title: Pascal-5i Dataset for Few-shot Object Segmentation
; Author: Mennatullah Siam
; Company: Huawei Technologies
; Date:   18 March 2020
;==========================================
"""
import os
import json
import glob
import h5py
from PIL import Image, ImageOps

from torchmeta.utils.data import Dataset, ClassDataset, CombinationMetaDataset
from torchvision.datasets.utils import list_dir, download_url
from torchmeta.datasets.utils import get_asset
import numpy as np

class Pascal5i(CombinationMetaDataset):
    """
    Pascal5i dataset [1]. A dataset for few-shot object segmentation supporting 4 folds
    each fold has 15 training classes and 5 testing classes.
    Using Preprocessed Masks from [2]

    Parameters
    ----------
    root : string
        Root directory where the dataset folder `omniglot` exists.

    num_classes_per_task : int
        Number of classes per tasks. This corresponds to "N" in "N-way"
        classification.

    meta_train : bool (default: `False`)
        Use the meta-train split of the dataset. If set to `True`, then the
        arguments `meta_val` and `meta_test` must be set to `False`. Exactly one
        of these three arguments must be set to `True`.

    meta_test : bool (default: `False`)
        Use the meta-test split of the dataset. If set to `True`, then the
        arguments `meta_train` and `meta_val` must be set to `False`. Exactly one
        of these three arguments must be set to `True`.

    meta_split : string in {'train', 'test'}, optional
        Name of the split to use. This overrides the arguments `meta_train`,
        and `meta_test` if all three are set to `False`.

    transform : callable, optional
        A function/transform that takes a `PIL` image, and returns a transformed
        version. See also `torchvision.transforms`.

    dataset_transform : callable, optional
        A function/transform that takes a dataset (ie. a task), and returns a
        transformed version of it. E.g. `torchmeta.transforms.ClassSplitter()`.

    class_augmentations : list of callable, optional
        A list of functions that augment the dataset with new classes. These classes
        are transformations of existing classes. E.g.
        `torchmeta.transforms.HorizontalFlip()`.

    download : bool (default: `False`)
        If `True`, downloads the zip files and processes the dataset in the root
        directory (under the `omniglot` folder). If the dataset is already
        available, this does not download/process the dataset again.

    fold : int (default: 0)
        Fold number ranges between 0-3 that controls training(15) and testing(5) classes.

    Notes
    -----
    Currently Only 1-way is supported

    References
    ----------
    .. [1] Shaban, Amirreza, et al. "One-shot learning for semantic segmentation."
            arXiv preprint arXiv:1709.03410 (2017).
    .. [2] Zhang, Chi, et al. "Canet: Class-agnostic segmentation networks with
            iterative refinement and attentive few-shot learning."
            Proceedings of the IEEE Conference on Computer Vision and Pattern Recognition. 2019.
    """
    def __init__(self, root, num_classes_per_task=None, meta_train=False,
                 meta_test=False, meta_split=None,
                 transform=None, target_transform=None,
                 dataset_transform=None, class_augmentations=None,
                 download=False, fold=0):

        dataset = Pascal5iClassDataset(root, meta_train=meta_train,
                                       meta_test=meta_test, transform=transform,
                                       meta_split=meta_split, class_augmentations=class_augmentations,
                                       download=download, fold=fold)

        super(Pascal5i, self).__init__(dataset, num_classes_per_task,
            target_transform=target_transform, dataset_transform=dataset_transform)

class Pascal5iClassDataset(ClassDataset):
    folder = 'pascal5i'

    downloads = [
    {
        'url' : 'http://host.robots.ox.ac.uk/pascal/VOC/voc2012/VOCtrainval_11-May-2012.tar',
        'filename' : 'VOCtrainval_11-May-2012.tar',
        'md5' : '6cd6e144f989b92b3379bac3b3de84fd'
    },
    {
        'url' : 'https://github.com/icoz69/CaNet/raw/master/Binary_map_aug.zip',
        'filename': 'Binary_map_aug.zip',
        'md5': None
    },
    {
        'url' : 'https://raw.github.com/NVIDIA/DIGITS/master/examples/semantic-segmentation/pascal-voc-classes.txt',
        'filename' : 'pascal-voc-classes.txt',
        'md5' : None
    }
    ]
    split_filename_labels = 'pascal-voc-classes.txt'

    def __init__(self, root, meta_train=False, meta_test=False,
                 meta_split=None, transform=None, class_augmentations=None,
                 download=False, fold=0):

        super(Pascal5iClassDataset, self).__init__(meta_train=meta_train,
            meta_val=False, meta_test=meta_test, meta_split=meta_split,
            class_augmentations=class_augmentations)

        self.root = os.path.join(os.path.expanduser(root), self.folder)
        self.transform = transform

        self.fold = fold

        self._data = None
        self._labels = None
        self._masks = None

        if download:
            self.download()

        self._num_classes = len(self.labels)

    def __getitem__(self, index):
        class_name = self.labels[index % self.num_classes]
        data, masks = self.data[0][class_name], self.data[1][class_name]
        transform = self.get_transform(index, self.transform)
        target_transform = self.get_target_transform(index)
        class_id = self.read_labels().index(class_name)

        return PascalDataset(index, (data, masks), class_id,
            transform=transform, target_transform=target_transform)

    @property
    def num_classes(self):
        return self._num_classes

    def load_dict_per_class(self):
        new_exist_class_list = {}

        if self.meta_split == 'train':
            fold_list=[0, 1, 2, 3]
            fold_list.remove(self.fold)
        else:
            fold_list = [self.fold]

        for fold in fold_list:
            f = open(os.path.join(self.root, 'Binary_map_aug', self.meta_split,
                                  'split%1d_%s.txt'%(fold, self.meta_split)))
            while True:
                item = f.readline()
                if item == '':
                    break
                img_name = item[:11]
                cat = int(item[13:15])
                if cat not in new_exist_class_list:
                    new_exist_class_list[cat] = []
                new_exist_class_list[cat].append(img_name)

        images = {}
        masks = {}
        classes_names = self.read_labels()

        for k, v in new_exist_class_list.items():
            cname = classes_names[k]
            for path in v:
                fname = os.path.join(self.root, 'VOCdevkit/VOC2012/JPEGImages', path + '.jpg')
                if cname not in images:
                    images[cname] = []
                images[cname].append(fname)
                fname = os.path.join(self.root, 'Binary_map_aug', self.meta_split, str(k),
                                               path + '.png')
                if cname not in masks:
                    masks[cname] = []
                masks[cname].append(fname)
        return images, masks

    @property
    def data(self):
        if self._data is None:
            self._data, self._masks = self.load_dict_per_class()
        return self._data, self._masks

    def read_labels(self, fold=None):
        labels = []
        if fold is not None:
            if self.meta_train:
                in_classes = set(range(21)) - \
                                set(range(fold*5+1, (fold+1)*5+1))
            else:
                in_classes = set(range(fold*5+1, (fold+1)*5+1))
        else:
            in_classes = set(range(21))

        with open(os.path.join(self.root, self.split_filename_labels), 'r') as f:
            for it, line in enumerate(f):
                if line.strip() == '':
                    break
                if it in in_classes:
                    labels.append(line.strip())
        return labels

    @property
    def labels(self):
        if self._labels is None:
           self._labels = self.read_labels(self.fold)
        return self._labels[1:]

    def download(self):
        import zipfile
        import tarfile
        import shutil

        for dload in self.downloads:
            filename = os.path.join(self.root, dload['filename'])
            if os.path.isfile(filename):
                continue

            download_url(dload['url'], self.root, dload['filename'], dload['md5'])

            if 'zip' in dload['filename']:
                with zipfile.ZipFile(filename, 'r') as f:
                    f.extractall(self.root)
            elif 'tar' in dload['filename']:
                with tarfile.open(filename, 'r') as f:
                    f.extractall(self.root)

class PascalDataset(Dataset):
    def __init__(self, index, data, class_id,
                 transform=None, target_transform=None):
        super(PascalDataset, self).__init__(index, transform=transform,
            target_transform=target_transform)
        self.data, self.masks = data
        self.class_id = class_id

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        image = Image.open(self.data[index])
        mask = Image.open(self.masks[index])
        target = self.class_id

        if self.transform is not None:
            image, mask = self.transform(image, mask)

        return (image, mask, target)



import os
import json
import glob
import h5py
from PIL import Image, ImageOps

from torchmeta.utils.data import Dataset, ClassDataset, CombinationMetaDataset
from torchvision.datasets.utils import list_dir, download_url
from torchmeta.datasets.utils import get_asset


class Omniglot(CombinationMetaDataset):
    """
    The Omniglot dataset [1]. A dataset of 1623 handwritten characters from 
    50 different alphabets. 

    Parameters
    ----------
    root : string
        Root directory where the dataset folder `omniglot` exists.

    num_classes_per_task : int
        Number of classes per tasks. This corresponds to "N" in "N-way" 
        classification.

    meta_train : bool (default: `False`)
        Use the meta-train split of the dataset. If set to `True`, then the
        arguments `meta_val` and `meta_test` must be set to `False`. Exactly one 
        of these three arguments must be set to `True`.

    meta_val : bool (default: `False`)
        Use the meta-validation split of the dataset. If set to `True`, then the 
        arguments `meta_train` and `meta_test` must be set to `False`. Exactly one 
        of these three arguments must be set to `True`.

    meta_test : bool (default: `False`)
        Use the meta-test split of the dataset. If set to `True`, then the 
        arguments `meta_train` and `meta_val` must be set to `False`. Exactly one 
        of these three arguments must be set to `True`.

    meta_split : string in {'train', 'val', 'test'}, optional
        Name of the split to use. This overrides the arguments `meta_train`, 
        `meta_val` and `meta_test` if all three are set to `False`.

    use_vinyals_split : bool (default: `True`)
        If set to `True`, the dataset uses the splits defined in [3]. If `False`, 
        then the meta-train split corresponds to `images_background`, and the 
        meta-test split corresponds to `images_evaluation` (raises an error when 
        calling the meta-validation split).

    transform : callable, optional
        A function/transform that takes a `PIL` image, and returns a transformed 
        version. See also `torchvision.transforms`.

    target_transform : callable, optional
        A function/transform that takes a target, and returns a transformed 
        version. See also `torchvision.transforms`.

    dataset_transform : callable, optional
        A function/transform that takes a dataset (ie. a task), and returns a 
        transformed version of it. E.g. `torchmeta.transforms.ClassSplitter()`.

    class_augmentations : list of callable, optional
        A list of functions that augment the dataset with new classes. These classes 
        are transformations of existing classes. E.g.
        `torchmeta.transforms.HorizontalFlip()`.

    download : bool (default: `False`)
        If `True`, downloads the zip files and processes the dataset in the root 
        directory (under the `omniglot` folder). If the dataset is already 
        available, this does not download/process the dataset again.

    Notes
    -----
    The dataset is downloaded from the original [Omniglot repository]
    (https://github.com/brendenlake/omniglot). The meta train/validation/test 
    splits used in [3] are taken from [this repository]
    (https://github.com/jakesnell/prototypical-networks). These splits are 
    over 1028/172/423 classes (characters).

    References
    ----------
    .. [1] Lake, B. M., Salakhutdinov, R., and Tenenbaum, J. B. (2015). Human-level 
           concept learning through probabilistic program induction. Science, 350(6266), 
           1332-1338 (http://www.sciencemag.org/content/350/6266/1332.short)

    .. [2] Lake, B. M., Salakhutdinov, R., and Tenenbaum, J. B. (2019). The Omniglot 
           Challenge: A 3-Year Progress Report (https://arxiv.org/abs/1902.03477)

    .. [3] Vinyals, O., Blundell, C., Lillicrap, T. and Wierstra, D. (2016). 
           Matching Networks for One Shot Learning. In Advances in Neural 
           Information Processing Systems (pp. 3630-3638) (https://arxiv.org/abs/1606.04080)
    """
    def __init__(self, root, num_classes_per_task=None, meta_train=False,
                 meta_val=False, meta_test=False, meta_split=None,
                 use_vinyals_split=True, transform=None, target_transform=None,
                 dataset_transform=None, class_augmentations=None, download=False):
        dataset = OmniglotClassDataset(root, meta_train=meta_train,
            meta_val=meta_val, meta_test=meta_test,
            use_vinyals_split=use_vinyals_split, transform=transform,
            meta_split=meta_split, class_augmentations=class_augmentations,
            download=download)
        super(Omniglot, self).__init__(dataset, num_classes_per_task,
            target_transform=target_transform, dataset_transform=dataset_transform)


class OmniglotClassDataset(ClassDataset):
    folder = 'omniglot'
    download_url_prefix = 'https://github.com/brendenlake/omniglot/raw/master/python'
    zips_md5 = {
        'images_background': '68d2efa1b9178cc56df9314c21c6e718',
        'images_evaluation': '6b91aef0f799c5bb55b94e3f2daec811'
    }

    filename = 'data.hdf5'
    filename_labels = '{0}{1}_labels.json'

    def __init__(self, root, meta_train=False, meta_val=False, meta_test=False,
                 meta_split=None, use_vinyals_split=True, transform=None,
                 class_augmentations=None, download=False):
        super(OmniglotClassDataset, self).__init__(meta_train=meta_train,
            meta_val=meta_val, meta_test=meta_test, meta_split=meta_split,
            class_augmentations=class_augmentations)

        if self.meta_val and (not use_vinyals_split):
            raise ValueError('Trying to use the meta-validation without the '
                'Vinyals split. You must set `use_vinyals_split=True` to use '
                'the meta-validation split.')

        self.root = os.path.join(os.path.expanduser(root), self.folder)
        self.use_vinyals_split = use_vinyals_split
        self.transform = transform

        self.split_filename = os.path.join(self.root, self.filename)
        self.split_filename_labels = os.path.join(self.root,
            self.filename_labels.format('vinyals_' if use_vinyals_split else '',
            self.meta_split))

        self._data = None
        self._labels = None

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError('Omniglot integrity check failed')
        self._num_classes = len(self.labels)

    def __getitem__(self, index):
        character_name = '/'.join(self.labels[index % self.num_classes])
        data = self.data[character_name]
        transform = self.get_transform(index, self.transform)
        target_transform = self.get_target_transform(index)

        return OmniglotDataset(index, data, character_name,
            transform=transform, target_transform=target_transform)

    @property
    def num_classes(self):
        return self._num_classes

    @property
    def data(self):
        if self._data is None:
            self._data = h5py.File(self.split_filename, 'r')
        return self._data

    @property
    def labels(self):
        if self._labels is None:
            with open(self.split_filename_labels, 'r') as f:
                self._labels = json.load(f)
        return self._labels

    def _check_integrity(self):
        return (os.path.isfile(self.split_filename)
            and os.path.isfile(self.split_filename_labels))

    def close(self):
        if self._data is not None:
            self._data.close()
            self._data = None

    def download(self):
        import zipfile
        import shutil

        if self._check_integrity():
            return

        for name in self.zips_md5:
            zip_filename = '{0}.zip'.format(name)
            filename = os.path.join(self.root, zip_filename)
            if os.path.isfile(filename):
                continue

            url = '{0}/{1}'.format(self.download_url_prefix, zip_filename)
            download_url(url, self.root, zip_filename, self.zips_md5[name])

            with zipfile.ZipFile(filename, 'r') as f:
                f.extractall(self.root)

        filename = os.path.join(self.root, self.filename)
        with h5py.File(filename, 'w') as f:
            for name in self.zips_md5:
                group = f.create_group(name)

                alphabets = list_dir(os.path.join(self.root, name))
                characters = [(name, alphabet, character) for alphabet in alphabets
                    for character in list_dir(os.path.join(self.root, name, alphabet))]

                split = 'train' if name == 'images_background' else 'test'
                labels_filename = os.path.join(self.root,
                    self.filename_labels.format('', split))
                with open(labels_filename, 'w') as f_labels:
                    labels = sorted(characters)
                    json.dump(labels, f_labels)

                for _, alphabet, character in characters:
                    filenames = glob.glob(os.path.join(self.root, name,
                        alphabet, character, '*.png'))
                    dataset = group.create_dataset('{0}/{1}'.format(alphabet,
                        character), (len(filenames), 105, 105), dtype='uint8')

                    for i, char_filename in enumerate(filenames):
                        image = Image.open(char_filename, mode='r').convert('L')
                        dataset[i] = ImageOps.invert(image)

                shutil.rmtree(os.path.join(self.root, name))

        for split in ['train', 'val', 'test']:
            filename = os.path.join(self.root, self.filename_labels.format(
                'vinyals_', split))
            data = get_asset(self.folder, '{0}.json'.format(split), dtype='json')

            with open(filename, 'w') as f:
                labels = sorted([('images_{0}'.format(name), alphabet, character)
                    for (name, alphabets) in data.items()
                    for (alphabet, characters) in alphabets.items()
                    for character in characters])
                json.dump(labels, f)


class OmniglotDataset(Dataset):
    def __init__(self, index, data, character_name,
                 transform=None, target_transform=None):
        super(OmniglotDataset, self).__init__(index, transform=transform,
                                              target_transform=target_transform)
        self.data = data
        self.character_name = character_name

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        image = Image.fromarray(self.data[index])
        target = self.character_name

        if self.transform is not None:
            image = self.transform(image)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return (image, target)



from collections import OrderedDict

from torch.utils.data import DataLoader
from torch.utils.data.dataloader import default_collate
from torch.utils.data.dataset import Dataset as TorchDataset

from torchmeta.utils.data.dataset import CombinationMetaDataset
from torchmeta.utils.data.sampler import (CombinationSequentialSampler,
                                          CombinationRandomSampler)

class BatchMetaCollate(object):

    def __init__(self, collate_fn):
        super().__init__()
        self.collate_fn = collate_fn

    def collate_task(self, task):
        if isinstance(task, TorchDataset):
            return self.collate_fn([task[idx] for idx in range(len(task))])
        elif isinstance(task, OrderedDict):
            return OrderedDict([(key, self.collate_task(subtask))
                for (key, subtask) in task.items()])
        else:
            raise NotImplementedError()

    def __call__(self, batch):
        return self.collate_fn([self.collate_task(task) for task in batch])

def no_collate(batch):
    return batch

class MetaDataLoader(DataLoader):
    def __init__(self, dataset, batch_size=1, shuffle=True, sampler=None,
                 batch_sampler=None, num_workers=0, collate_fn=None,
                 pin_memory=False, drop_last=False, timeout=0,
                 worker_init_fn=None):
        if collate_fn is None:
            collate_fn = no_collate

        if isinstance(dataset, CombinationMetaDataset) and (sampler is None):
            if shuffle:
                sampler = CombinationRandomSampler(dataset)
            else:
                sampler = CombinationSequentialSampler(dataset)
            shuffle = False

        super(MetaDataLoader, self).__init__(dataset, batch_size=batch_size,
            shuffle=shuffle, sampler=sampler, batch_sampler=batch_sampler,
            num_workers=num_workers, collate_fn=collate_fn,
            pin_memory=pin_memory, drop_last=drop_last, timeout=timeout,
            worker_init_fn=worker_init_fn)


class BatchMetaDataLoader(MetaDataLoader):
    def __init__(self, dataset, batch_size=1, shuffle=True, sampler=None, num_workers=0,
                 pin_memory=False, drop_last=False, timeout=0, worker_init_fn=None):
        collate_fn = BatchMetaCollate(default_collate)

        super(BatchMetaDataLoader, self).__init__(dataset,
            batch_size=batch_size, shuffle=shuffle, sampler=sampler,
            batch_sampler=None, num_workers=num_workers,
            collate_fn=collate_fn, pin_memory=pin_memory, drop_last=drop_last,
            timeout=timeout, worker_init_fn=worker_init_fn)



import torch

from collections import OrderedDict
from torchmeta.modules import MetaModule


def gradient_update_parameters(model,
                               loss,
                               params=None,
                               step_size=0.5,
                               first_order=False):
    """Update of the meta-parameters with one step of gradient descent on the
    loss function.

    Parameters
    ----------
    model : `torchmeta.modules.MetaModule` instance
        The model.

    loss : `torch.Tensor` instance
        The value of the inner-loss. This is the result of the training dataset
        through the loss function.

    params : `collections.OrderedDict` instance, optional
        Dictionary containing the meta-parameters of the model. If `None`, then
        the values stored in `model.meta_named_parameters()` are used. This is
        useful for running multiple steps of gradient descent as the inner-loop.

    step_size : int, `torch.Tensor`, or `collections.OrderedDict` instance (default: 0.5)
        The step size in the gradient update. If an `OrderedDict`, then the
        keys must match the keys in `params`.

    first_order : bool (default: `False`)
        If `True`, then the first order approximation of MAML is used.

    Returns
    -------
    updated_params : `collections.OrderedDict` instance
        Dictionary containing the updated meta-parameters of the model, with one
        gradient update wrt. the inner-loss.
    """
    if not isinstance(model, MetaModule):
        raise ValueError('The model must be an instance of `torchmeta.modules.'
                         'MetaModule`, got `{0}`'.format(type(model)))

    if params is None:
        params = OrderedDict(model.meta_named_parameters())

    grads = torch.autograd.grad(loss,
                                params.values(),
                                create_graph=not first_order)

    updated_params = OrderedDict()

    if isinstance(step_size, (dict, OrderedDict)):
        for (name, param), grad in zip(params.items(), grads):
            updated_params[name] = param - step_size[name] * grad

    else:
        for (name, param), grad in zip(params.items(), grads):
            updated_params[name] = param - step_size * grad

    return updated_params



from torchvision.transforms import Compose, Resize, ToTensor
import PIL

class SegmentationPairTransform(object):
    def __init__(self, target_size):
        self.image_transform = Compose([Resize((target_size, target_size)), ToTensor()])
        self.mask_transform = Compose([Resize((target_size, target_size),
                                               interpolation=PIL.Image.NEAREST),
                                       ToTensor()])

    def __call__(self, image, mask):
        image = self.image_transform(image)
        mask = self.mask_transform(mask)
        return image, mask

class TargetTransform(object):
    def __call__(self, target):
        raise NotImplementedError()

    def __repr__(self):
        return str(self.__class__.__name__)


class DefaultTargetTransform(TargetTransform):
    def __init__(self, class_augmentations):
        super(DefaultTargetTransform, self).__init__()
        self.class_augmentations = class_augmentations

        self._augmentations = dict((augmentation, i + 1)
            for (i, augmentation) in enumerate(class_augmentations))
        self._augmentations[None] = 0

    def __call__(self, target):
        assert isinstance(target, tuple) and len(target) == 2
        label, augmentation = target
        return (label, self._augmentations[augmentation])



import torch.nn as nn
import torch.nn.functional as F

from collections import OrderedDict
from torchmeta.modules.module import MetaModule

class MetaLayerNorm(nn.LayerNorm, MetaModule):
    __doc__ = nn.LayerNorm.__doc__

    def forward(self, input, params=None):
        if params is None:
            params = OrderedDict(self.named_parameters())
        weight = params.get('weight', None)
        bias = params.get('bias', None)
        return F.layer_norm(
            input, self.normalized_shape, weight, bias, self.eps)



import torch.nn as nn

from torchmeta.modules.module import MetaModule


class MetaSequential(nn.Sequential, MetaModule):
    __doc__ = nn.Sequential.__doc__

    def forward(self, input, params=None):
        for name, module in self._modules.items():
            if isinstance(module, MetaModule):
                input = module(input, params=self.get_subdict(params, name))
            elif isinstance(module, nn.Module):
                input = module(input)
            else:
                raise TypeError('The module must be either a torch module '
                    '(inheriting from `nn.Module`), or a `MetaModule`. '
                    'Got type: `{0}`'.format(type(module)))
        return input



import pytest

import numpy as np
import torch
import torch.nn as nn
from collections import OrderedDict

from torchmeta.modules import MetaModule
from torchmeta.modules.conv import MetaConv1d, MetaConv2d, MetaConv3d


@pytest.mark.parametrize('bias', [True, False])
def test_metaconv1d(bias):
    meta_model = MetaConv1d(2, 3, kernel_size=5, bias=bias)
    model = nn.Conv1d(2, 3, kernel_size=5, bias=bias)

    assert isinstance(meta_model, MetaModule)
    assert isinstance(meta_model, nn.Conv1d)

    # Set same weights for both models
    weight = torch.randn(3, 2, 5)
    meta_model.weight.data.copy_(weight)
    model.weight.data.copy_(weight)

    if bias:
        bias = torch.randn(3)
        meta_model.bias.data.copy_(bias)
        model.bias.data.copy_(bias)

    inputs = torch.randn(7, 2, 11)

    outputs_torchmeta = meta_model(inputs, params=None)
    outputs_nn = model(inputs)

    np.testing.assert_equal(outputs_torchmeta.detach().numpy(),
                            outputs_nn.detach().numpy())


@pytest.mark.parametrize('bias', [True, False])
def test_metaconv1d_params(bias):
    meta_model = MetaConv1d(2, 3, kernel_size=5, bias=bias)
    model = nn.Conv1d(2, 3, kernel_size=5, bias=bias)

    params = OrderedDict()
    params['weight'] = torch.randn(3, 2, 5)
    model.weight.data.copy_(params['weight'])

    if bias:
        params['bias'] = torch.randn(3)
        model.bias.data.copy_(params['bias'])

    inputs = torch.randn(7, 2, 11)

    outputs_torchmeta = meta_model(inputs, params=params)
    outputs_nn = model(inputs)

    np.testing.assert_equal(outputs_torchmeta.detach().numpy(),
                            outputs_nn.detach().numpy())


@pytest.mark.parametrize('bias', [True, False])
def test_metaconv2d(bias):
    meta_model = MetaConv2d(2, 3, kernel_size=(5, 7), bias=bias)
    model = nn.Conv2d(2, 3, kernel_size=(5, 7), bias=bias)

    assert isinstance(meta_model, MetaModule)
    assert isinstance(meta_model, nn.Conv2d)

    # Set same weights for both models
    weight = torch.randn(3, 2, 5, 7)
    meta_model.weight.data.copy_(weight)
    model.weight.data.copy_(weight)

    if bias:
        bias = torch.randn(3)
        meta_model.bias.data.copy_(bias)
        model.bias.data.copy_(bias)

    inputs = torch.randn(11, 2, 13, 17)

    outputs_torchmeta = meta_model(inputs, params=None)
    outputs_nn = model(inputs)

    np.testing.assert_equal(outputs_torchmeta.detach().numpy(),
                            outputs_nn.detach().numpy())


@pytest.mark.parametrize('bias', [True, False])
def test_metaconv2d_params(bias):
    meta_model = MetaConv2d(2, 3, kernel_size=(5, 7), bias=bias)
    model = nn.Conv2d(2, 3, kernel_size=(5, 7), bias=bias)

    params = OrderedDict()
    params['weight'] = torch.randn(3, 2, 5, 7)
    model.weight.data.copy_(params['weight'])

    if bias:
        params['bias'] = torch.randn(3)
        model.bias.data.copy_(params['bias'])

    inputs = torch.randn(11, 2, 13, 17)

    outputs_torchmeta = meta_model(inputs, params=params)
    outputs_nn = model(inputs)

    np.testing.assert_equal(outputs_torchmeta.detach().numpy(),
                            outputs_nn.detach().numpy())


@pytest.mark.parametrize('bias', [True, False])
def test_metaconv3d(bias):
    meta_model = MetaConv3d(2, 3, kernel_size=(5, 7, 11), bias=bias)
    model = nn.Conv3d(2, 3, kernel_size=(5, 7, 11), bias=bias)

    assert isinstance(meta_model, MetaModule)
    assert isinstance(meta_model, nn.Conv3d)

    # Set same weights for both models
    weight = torch.randn(3, 2, 5, 7, 11)
    meta_model.weight.data.copy_(weight)
    model.weight.data.copy_(weight)

    if bias:
        bias = torch.randn(3)
        meta_model.bias.data.copy_(bias)
        model.bias.data.copy_(bias)

    inputs = torch.randn(13, 2, 17, 23, 29)

    outputs_torchmeta = meta_model(inputs, params=None)
    outputs_nn = model(inputs)

    np.testing.assert_equal(outputs_torchmeta.detach().numpy(),
                            outputs_nn.detach().numpy())


@pytest.mark.parametrize('bias', [True, False])
def test_metaconv3d_params(bias):
    meta_model = MetaConv3d(2, 3, kernel_size=(5, 7, 11), bias=bias)
    model = nn.Conv3d(2, 3, kernel_size=(5, 7, 11), bias=bias)

    params = OrderedDict()
    params['weight'] = torch.randn(3, 2, 5, 7, 11)
    model.weight.data.copy_(params['weight'])

    if bias:
        params['bias'] = torch.randn(3)
        model.bias.data.copy_(params['bias'])

    inputs = torch.randn(13, 2, 17, 23, 29)

    outputs_torchmeta = meta_model(inputs, params=params)
    outputs_nn = model(inputs)

    np.testing.assert_equal(outputs_torchmeta.detach().numpy(),
                            outputs_nn.detach().numpy())



from torchmeta.toy.harmonic import Harmonic
from torchmeta.toy.sinusoid import Sinusoid
from torchmeta.toy.sinusoid_line import SinusoidAndLine

from torchmeta.toy import helpers

__all__ = ['Harmonic', 'Sinusoid', 'SinusoidAndLine', 'helpers']



from torchmeta.datasets.triplemnist import TripleMNIST
from torchmeta.datasets.doublemnist import DoubleMNIST
from torchmeta.datasets.cub import CUB
from torchmeta.datasets.cifar100 import CIFARFS, FC100
from torchmeta.datasets.miniimagenet import MiniImagenet
from torchmeta.datasets.omniglot import Omniglot
from torchmeta.datasets.tieredimagenet import TieredImagenet
from torchmeta.datasets.tcga import TCGA
from torchmeta.datasets.pascal5i import Pascal5i

from torchmeta.datasets import helpers

__all__ = [
    'TCGA',
    'Omniglot',
    'MiniImagenet',
    'TieredImagenet',
    'CIFARFS',
    'FC100',
    'CUB',
    'DoubleMNIST',
    'TripleMNIST',
    'Pascal5i',
    'helpers'
]



import os
import json

from torchmeta.datasets.cifar100.base import CIFAR100ClassDataset
from torchmeta.datasets.utils import get_asset
from torchmeta.utils.data import ClassDataset, CombinationMetaDataset


class CIFARFS(CombinationMetaDataset):
    """
    The CIFAR-FS dataset, introduced in [1]. This dataset contains
    images of 100 different classes from the CIFAR100 dataset [2].

    Parameters
    ----------
    root : string
        Root directory where the dataset folder `cifar100` exists.

    num_classes_per_task : int
        Number of classes per tasks. This corresponds to `N` in `N-way` 
        classification.

    meta_train : bool (default: `False`)
        Use the meta-train split of the dataset. If set to `True`, then the
        arguments `meta_val` and `meta_test` must be set to `False`. Exactly one 
        of these three arguments must be set to `True`.

    meta_val : bool (default: `False`)
        Use the meta-validation split of the dataset. If set to `True`, then the 
        arguments `meta_train` and `meta_test` must be set to `False`. Exactly one 
        of these three arguments must be set to `True`.

    meta_test : bool (default: `False`)
        Use the meta-test split of the dataset. If set to `True`, then the 
        arguments `meta_train` and `meta_val` must be set to `False`. Exactly one 
        of these three arguments must be set to `True`.

    meta_split : string in {'train', 'val', 'test'}, optional
        Name of the split to use. This overrides the arguments `meta_train`, 
        `meta_val` and `meta_test` if all three are set to `False`.

    transform : callable, optional
        A function/transform that takes a `PIL` image, and returns a transformed 
        version. See also `torchvision.transforms`.

    target_transform : callable, optional
        A function/transform that takes a target, and returns a transformed 
        version. See also `torchvision.transforms`.

    dataset_transform : callable, optional
        A function/transform that takes a dataset (ie. a task), and returns a 
        transformed version of it. E.g. `transforms.ClassSplitter()`.

    class_augmentations : list of callable, optional
        A list of functions that augment the dataset with new classes. These classes 
        are transformations of existing classes. E.g. `transforms.HorizontalFlip()`.

    download : bool (default: `False`)
        If `True`, downloads the pickle files and processes the dataset in the root 
        directory (under the `cifar100` folder). If the dataset is already 
        available, this does not download/process the dataset again.

    Notes
    -----
    The meta train/validation/test splits are over 64/16/20 classes from the
    CIFAR100 dataset.

    References
    ----------
    .. [1] Bertinetto L., Henriques J. F., Torr P. H.S., Vedaldi A. (2019).
           Meta-learning with differentiable closed-form solvers. In International
           Conference on Learning Representations (https://arxiv.org/abs/1805.08136)

    .. [2] Krizhevsky A. (2009). Learning Multiple Layers of Features from Tiny
           Images. (https://www.cs.toronto.edu/~kriz/learning-features-2009-TR.pdf)
    """
    def __init__(self, root, num_classes_per_task=None, meta_train=False,
                 meta_val=False, meta_test=False, meta_split=None,
                 transform=None, target_transform=None, dataset_transform=None,
                 class_augmentations=None, download=False):
        dataset = CIFARFSClassDataset(root, meta_train=meta_train,
            meta_val=meta_val, meta_test=meta_test, meta_split=meta_split,
            transform=transform, class_augmentations=class_augmentations,
            download=download)
        super(CIFARFS, self).__init__(dataset, num_classes_per_task,
            target_transform=target_transform, dataset_transform=dataset_transform)


class CIFARFSClassDataset(CIFAR100ClassDataset):
    subfolder = 'cifar-fs'

    def __init__(self, root, meta_train=False, meta_val=False, meta_test=False,
                 meta_split=None, transform=None, class_augmentations=None,
                 download=False):
        super(CIFARFSClassDataset, self).__init__(root, meta_train=meta_train,
            meta_val=meta_val, meta_test=meta_test, meta_split=meta_split,
            transform=transform, class_augmentations=class_augmentations,
            download=download)

    def download(self):
        if self._check_integrity():
            return
        super(CIFARFSClassDataset, self).download()

        subfolder = os.path.join(self.root, self.subfolder)
        if not os.path.exists(subfolder):
            os.makedirs(subfolder)

        for split in ['train', 'val', 'test']:
            split_filename_labels = os.path.join(subfolder,
                self.filename_labels.format(split))
            if os.path.isfile(split_filename_labels):
                continue

            data = get_asset(self.folder, self.subfolder,
                '{0}.json'.format(split), dtype='json')
            with open(split_filename_labels, 'w') as f:
                json.dump(data, f)



import pytest

import numpy as np
import torch
import torch.nn as nn
from collections import OrderedDict

from torchmeta.modules import MetaSequential, MetaModule, MetaLinear


def test_metasequential():
    meta_model = MetaSequential(
        nn.Linear(2, 3, bias=True),
        nn.ReLU(),
        MetaLinear(3, 5, bias=True))
    model = nn.Sequential(
        nn.Linear(2, 3, bias=True),
        nn.ReLU(),
        nn.Linear(3, 5, bias=True))

    assert isinstance(meta_model, MetaModule)
    assert isinstance(meta_model, nn.Sequential)

    params = OrderedDict(meta_model.meta_named_parameters())
    assert set(params.keys()) == set(['2.weight', '2.bias'])

    # Set same weights for both models
    weight0 = torch.randn(3, 2)
    meta_model[0].weight.data.copy_(weight0)
    model[0].weight.data.copy_(weight0)

    bias0 = torch.randn(3)
    meta_model[0].bias.data.copy_(bias0)
    model[0].bias.data.copy_(bias0)

    weight2 = torch.randn(5, 3)
    meta_model[2].weight.data.copy_(weight2)
    model[2].weight.data.copy_(weight2)

    bias2 = torch.randn(5)
    meta_model[2].bias.data.copy_(bias2)
    model[2].bias.data.copy_(bias2)

    inputs = torch.randn(5, 2)

    outputs_torchmeta = meta_model(inputs, params=None)
    outputs_nn = model(inputs)

    np.testing.assert_equal(outputs_torchmeta.detach().numpy(),
                            outputs_nn.detach().numpy())


def test_metasequential_params():
    meta_model = MetaSequential(
        nn.Linear(2, 3, bias=True),
        nn.ReLU(),
        MetaLinear(3, 5, bias=True))
    model = nn.Sequential(
        nn.Linear(2, 3, bias=True),
        nn.ReLU(),
        nn.Linear(3, 5, bias=True))

    # Set same weights for both models (first layer)
    weight0 = torch.randn(3, 2)
    meta_model[0].weight.data.copy_(weight0)
    model[0].weight.data.copy_(weight0)

    bias0 = torch.randn(3)
    meta_model[0].bias.data.copy_(bias0)
    model[0].bias.data.copy_(bias0)

    params = OrderedDict()
    params['2.weight'] = torch.randn(5, 3)
    model[2].weight.data.copy_(params['2.weight'])

    params['2.bias'] = torch.randn(5)
    model[2].bias.data.copy_(params['2.bias'])

    inputs = torch.randn(5, 2)

    outputs_torchmeta = meta_model(inputs, params=params)
    outputs_nn = model(inputs)

    np.testing.assert_equal(outputs_torchmeta.detach().numpy(),
                            outputs_nn.detach().numpy())



from torch.utils.data import ConcatDataset, Subset
from torch.utils.data import Dataset as Dataset_
from torchvision.transforms import Compose

__all__ = ['Dataset', 'Task', 'ConcatTask', 'SubsetTask']


class Dataset(Dataset_):
    def __init__(self, index, transform=None, target_transform=None):
        self.index = index
        self.transform = transform
        self.target_transform = target_transform

    def target_transform_append(self, transform):
        if transform is None:
            return
        if self.target_transform is None:
            self.target_transform = transform
        else:
            self.target_transform = Compose([self.target_transform, transform])

    def __hash__(self):
        return hash(self.index)


class Task(Dataset):
    """Base class for a classification task.

    Parameters
    ----------
    num_classes : int
        Number of classes for the classification task.
    """
    def __init__(self, index, num_classes,
                 transform=None, target_transform=None):
        super(Task, self).__init__(index, transform=transform,
                                   target_transform=target_transform)
        self.num_classes = num_classes


class ConcatTask(Task, ConcatDataset):
    def __init__(self, datasets, num_classes, target_transform=None):
        index = tuple(task.index for task in datasets)
        Task.__init__(self, index, num_classes)
        ConcatDataset.__init__(self, datasets)
        for task in self.datasets:
            task.target_transform_append(target_transform)

    def __getitem__(self, index):
        return ConcatDataset.__getitem__(self, index)


class SubsetTask(Task, Subset):
    def __init__(self, dataset, indices, num_classes=None,
                 target_transform=None):
        if num_classes is None:
            num_classes = dataset.num_classes
        Task.__init__(self, dataset.index, num_classes)
        Subset.__init__(self, dataset, indices)
        self.dataset.target_transform_append(target_transform)

    def __getitem__(self, index):
        return Subset.__getitem__(self, index)

    def __hash__(self):
        return hash((self.index, tuple(self.indices)))



import torch
from torchmeta.transforms.utils import apply_wrapper
from collections import defaultdict

from torchmeta.transforms.target_transforms import TargetTransform


class Categorical(TargetTransform):
    """Target transform to return labels in `[0, num_classes)`.

    Parameters
    ----------
    num_classes : int, optional
        Number of classes. If `None`, then the number of classes is inferred
        from the number of individual labels encountered.

    Examples
    --------
    >>> dataset = Omniglot('data', num_classes_per_task=5, meta_train=True)
    >>> task = dataset.sample_task()
    >>> task[0]
    (<PIL.Image.Image image mode=L size=105x105 at 0x11EC797F0>,
    ('images_evaluation/Glagolitic/character12', None))

    >>> dataset = Omniglot('data', num_classes_per_task=5, meta_train=True,
    ... target_transform=Categorical(5))
    >>> task = dataset.sample_task()
    >>> task[0]
    (<PIL.Image.Image image mode=L size=105x105 at 0x11ED3F668>, 2)
    """
    def __init__(self, num_classes=None):
        super(Categorical, self).__init__()
        self.num_classes = num_classes
        self._classes = None
        self._labels = None

    def reset(self):
        self._classes = None
        self._labels = None

    @property
    def classes(self):
        if self._classes is None:
            self._classes = defaultdict(None)
            if self.num_classes is None:
                default_factory = lambda: len(self._classes)
            else:
                default_factory = lambda: self.labels[len(self._classes)]
            self._classes.default_factory = default_factory
        if (self.num_classes is not None) and (len(self._classes) > self.num_classes):
            raise ValueError('The number of individual labels ({0}) is greater '
                'than the number of classes defined by `num_classes` '
                '({1}).'.format(len(self._classes), self.num_classes))
        return self._classes

    @property
    def labels(self):
        if (self._labels is None) and (self.num_classes is not None):
            # TODO: Replace torch.randperm with seed-friendly counterpart
            self._labels = torch.randperm(self.num_classes).tolist()
        return self._labels

    def __call__(self, target):
        return self.classes[target]

    def __repr__(self):
        return '{0}({1})'.format(self.__class__.__name__, self.num_classes or '')


class FixedCategory(object):
    def __init__(self, transform=None):
        self.transform = transform

    def __call__(self, index):
        return (index, self.transform)

    def __repr__(self):
        return ('{0}({1})'.format(self.__class__.__name__, self.transform))



import pytest

import numpy as np
import torch

from torchmeta.utils.prototype import get_num_samples, get_prototypes, prototypical_loss


@pytest.mark.parametrize('dtype', [None, torch.float32])
def test_get_num_samples(dtype):
    # Numpy
    num_classes = 3
    targets_np = np.random.randint(0, num_classes, size=(2, 5))

    # PyTorch
    targets_th = torch.as_tensor(targets_np)
    num_samples_th = get_num_samples(targets_th, num_classes, dtype=dtype)

    num_samples_np = np.zeros((2, num_classes), dtype=np.int_)
    for i in range(2):
        for j in range(5):
            num_samples_np[i, targets_np[i, j]] += 1

    assert num_samples_th.shape == (2, num_classes)
    if dtype is not None:
        assert num_samples_th.dtype == dtype
    np.testing.assert_equal(num_samples_th.numpy(), num_samples_np)


def test_get_prototypes():
    # Numpy
    num_classes = 3
    embeddings_np = np.random.rand(2, 5, 7).astype(np.float32)
    targets_np = np.random.randint(0, num_classes, size=(2, 5))

    # PyTorch
    embeddings_th = torch.as_tensor(embeddings_np)
    targets_th = torch.as_tensor(targets_np)
    prototypes_th = get_prototypes(embeddings_th, targets_th, num_classes)

    assert prototypes_th.shape == (2, num_classes, 7)
    assert prototypes_th.dtype == embeddings_th.dtype

    prototypes_np = np.zeros((2, num_classes, 7), dtype=np.float32)
    num_samples_np = np.zeros((2, num_classes), dtype=np.int_)
    for i in range(2):
        for j in range(5):
            num_samples_np[i, targets_np[i, j]] += 1
            for k in range(7):
                prototypes_np[i, targets_np[i, j], k] += embeddings_np[i, j, k]

    for i in range(2):
        for j in range(num_classes):
            for k in range(7):
                prototypes_np[i, j, k] /= max(num_samples_np[i, j], 1)

    np.testing.assert_allclose(prototypes_th.detach().numpy(), prototypes_np)



import random
import warnings
from itertools import combinations
from torch.utils.data.sampler import SequentialSampler, RandomSampler

from torchmeta.utils.data.dataset import CombinationMetaDataset

__all__ = ['CombinationSequentialSampler', 'CombinationRandomSampler']


class CombinationSequentialSampler(SequentialSampler):
    def __init__(self, data_source):
        if not isinstance(data_source, CombinationMetaDataset):
            raise TypeError('Expected `data_source` to be an instance of '
                            '`CombinationMetaDataset`, but found '
                            '{0}'.format(type(data_source)))
        super(CombinationSequentialSampler, self).__init__(data_source)

    def __iter__(self):
        num_classes = len(self.data_source.dataset)
        num_classes_per_task = self.data_source.num_classes_per_task
        return combinations(range(num_classes), num_classes_per_task)


class CombinationRandomSampler(RandomSampler):
    def __init__(self, data_source):
        if not isinstance(data_source, CombinationMetaDataset):
            raise TypeError('Expected `data_source` to be an instance of '
                            '`CombinationMetaDataset`, but found '
                            '{0}'.format(type(data_source)))
        # Temporarily disable the warning if the length of the length of the 
        # dataset exceeds the machine precision. This avoids getting this
        # warning shown with MetaDataLoader, even though MetaDataLoader itself
        # does not use the length of the dataset.
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            super(CombinationRandomSampler, self).__init__(data_source,
                                                           replacement=True)

    def __iter__(self):
        num_classes = len(self.data_source.dataset)
        num_classes_per_task = self.data_source.num_classes_per_task
        for _ in combinations(range(num_classes), num_classes_per_task):
            yield tuple(random.sample(range(num_classes), num_classes_per_task))



import pytest

import numpy as np
from collections import OrderedDict

from torchmeta.utils.data import Task, MetaDataset
from torchmeta.toy import Sinusoid, Harmonic, SinusoidAndLine
from torchmeta.toy import helpers


@pytest.mark.parametrize('dataset_class',
    [Sinusoid, Harmonic, SinusoidAndLine])
def test_toy_meta_dataset(dataset_class):
    dataset = dataset_class(10, num_tasks=1000, noise_std=None)

    assert isinstance(dataset, MetaDataset)
    assert len(dataset) == 1000


@pytest.mark.parametrize('dataset_class',
    [Sinusoid, Harmonic, SinusoidAndLine])
def test_toy_task(dataset_class):
    dataset = dataset_class(10, num_tasks=1000, noise_std=None)
    task = dataset[0]

    assert isinstance(task, Task)
    assert len(task) == 10


@pytest.mark.parametrize('dataset_class',
    [Sinusoid, Harmonic, SinusoidAndLine])
def test_toy_sample(dataset_class):
    dataset = dataset_class(10, num_tasks=1000, noise_std=None)
    task = dataset[0]
    input, target = task[0]

    assert isinstance(input, np.ndarray)
    assert isinstance(target, np.ndarray)
    assert input.shape == (1,)
    assert target.shape == (1,)


@pytest.mark.parametrize('name,dataset_class',
    [('sinusoid', Sinusoid), ('harmonic', Harmonic)])
def test_toy_helpers(name, dataset_class):
    dataset_fn = getattr(helpers, name)
    dataset = dataset_fn(shots=5, test_shots=15)
    assert isinstance(dataset, dataset_class)

    task = dataset[0]
    assert isinstance(task, OrderedDict)
    assert 'train' in task
    assert 'test' in task

    train, test = task['train'], task['test']
    assert isinstance(train, Task)
    assert isinstance(test, Task)
    assert len(train) == 5
    assert len(test) == 15



import pytest

import numpy as np
from collections import OrderedDict
from ordered_set import OrderedSet

from torchmeta.transforms.splitters import ClassSplitter
from torchmeta.toy import Sinusoid
from torchmeta.utils.data import Task

def test_seed_class_splitter():
    dataset_transform = ClassSplitter(shuffle=True,
        num_train_per_class=5, num_test_per_class=5)
    dataset = Sinusoid(10, num_tasks=1000, noise_std=0.1,
        dataset_transform=dataset_transform)
    dataset.seed(1)

    expected_train_inputs = np.array([-2.03870077,  0.09898378,  3.75388738,  1.08565437, -1.56211897])
    expected_train_targets = np.array([-0.1031986 , -1.61885041,  0.91773121, -0.00309463, -1.37650356])

    expected_test_inputs = np.array([ 4.62078213, -2.48340416,  0.32922559,  0.76977846, -3.15504396])
    expected_test_targets = np.array([-0.9346262 ,  0.73113509, -1.52508997, -0.4698061 ,  1.86656819])

    task = dataset[0]
    train_dataset, test_dataset = task['train'], task['test']

    assert len(train_dataset) == 5
    assert len(test_dataset) == 5

    for i, (train_input, train_target) in enumerate(train_dataset):
        assert np.isclose(train_input, expected_train_inputs[i])
        assert np.isclose(train_target, expected_train_targets[i])

    for i, (test_input, test_target) in enumerate(test_dataset):
        assert np.isclose(test_input, expected_test_inputs[i])
        assert np.isclose(test_target, expected_test_targets[i])

def test_class_splitter_for_fold_overlaps():
    class DemoTask(Task):
        def __init__(self):
            super(DemoTask, self).__init__(index=0, num_classes=None)
            self._inputs = np.arange(10)

        def __len__(self):
            return len(self._inputs)

        def __getitem__(self, index):
            return self._inputs[index]

    splitter = ClassSplitter(shuffle=True, num_train_per_class=5, num_test_per_class=5)
    task = DemoTask()

    all_train_samples = list()
    all_test_samples = list()

    # split task ten times into train and test
    for i in range(10):
        tasks_split = splitter(task)
        train_task = tasks_split["train"]
        test_task = tasks_split["test"]

        train_samples = set([train_task[i] for i in range(len(train_task))])
        test_samples = set([test_task[i] for i in range(len(train_task))])

        # no overlap between train and test splits at single split
        assert len(train_samples.intersection(test_samples)) == 0

        all_train_samples.append(train_samples)
        all_train_samples.append(train_samples)

    # gather unique samples from multiple splits
    samples_in_all_train_splits = OrderedSet().union(*all_train_samples)
    samples_in_all_test_splits = OrderedSet().union(*all_test_samples)

    # no overlap between train and test splits at multiple splits
    assert len(samples_in_all_test_splits.intersection(samples_in_all_train_splits)) == 0



import pytest

import numpy as np
import torch
import torch.nn as nn
from collections import OrderedDict

from torchmeta.modules import MetaModule
from torchmeta.modules.linear import MetaLinear, MetaBilinear


@pytest.mark.parametrize('bias', [True, False])
def test_metalinear(bias):
    meta_model = MetaLinear(2, 3, bias=bias)
    model = nn.Linear(2, 3, bias=bias)

    assert isinstance(meta_model, MetaModule)
    assert isinstance(meta_model, nn.Linear)

    # Set same weights for both models
    weight = torch.randn(3, 2)
    meta_model.weight.data.copy_(weight)
    model.weight.data.copy_(weight)

    if bias:
        bias = torch.randn(3)
        meta_model.bias.data.copy_(bias)
        model.bias.data.copy_(bias)

    inputs = torch.randn(5, 2)

    outputs_torchmeta = meta_model(inputs, params=None)
    outputs_nn = model(inputs)

    np.testing.assert_equal(outputs_torchmeta.detach().numpy(),
                            outputs_nn.detach().numpy())


@pytest.mark.parametrize('bias', [True, False])
def test_metalinear_params(bias):
    meta_model = MetaLinear(2, 3, bias=bias)
    model = nn.Linear(2, 3, bias=bias)

    params = OrderedDict()
    params['weight'] = torch.randn(3, 2)
    model.weight.data.copy_(params['weight'])

    if bias:
        params['bias'] = torch.randn(3)
        model.bias.data.copy_(params['bias'])

    inputs = torch.randn(5, 2)

    outputs_torchmeta = meta_model(inputs, params=params)
    outputs_nn = model(inputs)

    np.testing.assert_equal(outputs_torchmeta.detach().numpy(),
                            outputs_nn.detach().numpy())


@pytest.mark.parametrize('bias', [True, False])
def test_metabilinear(bias):
    meta_model = MetaBilinear(2, 3, 5, bias=bias)
    model = nn.Bilinear(2, 3, 5, bias=bias)

    assert isinstance(meta_model, MetaModule)
    assert isinstance(meta_model, nn.Bilinear)

    # Set same weights for both models
    weight = torch.randn(5, 2, 3)
    meta_model.weight.data.copy_(weight)
    model.weight.data.copy_(weight)

    if bias:
        bias = torch.randn(5)
        meta_model.bias.data.copy_(bias)
        model.bias.data.copy_(bias)

    inputs1 = torch.randn(7, 2)
    inputs2 = torch.randn(7, 3)

    outputs_torchmeta = meta_model(inputs1, inputs2, params=None)
    outputs_nn = model(inputs1, inputs2)

    np.testing.assert_equal(outputs_torchmeta.detach().numpy(),
                            outputs_nn.detach().numpy())


@pytest.mark.parametrize('bias', [True, False])
def test_metabilinear_params(bias):
    meta_model = MetaBilinear(2, 3, 5, bias=bias)
    model = nn.Bilinear(2, 3, 5, bias=bias)

    params = OrderedDict()
    params['weight'] = torch.randn(5, 2, 3)
    model.weight.data.copy_(params['weight'])

    if bias:
        params['bias'] = torch.randn(5)
        model.bias.data.copy_(params['bias'])

    inputs1 = torch.randn(7, 2)
    inputs2 = torch.randn(7, 3)

    outputs_torchmeta = meta_model(inputs1, inputs2, params=params)
    outputs_nn = model(inputs1, inputs2)

    np.testing.assert_equal(outputs_torchmeta.detach().numpy(),
                            outputs_nn.detach().numpy())



import torch.nn as nn
import torch.nn.functional as F

from collections import OrderedDict
from torchmeta.modules.module import MetaModule

class MetaEmbedding(nn.Embedding, MetaModule):
    __doc__ = nn.Embedding.__doc__

    def forward(self, input, params=None):
        if params is None:
            params = OrderedDict(self.named_parameters())
        return F.embedding(
            input, params['weight'], self.padding_idx, self.max_norm,
            self.norm_type, self.scale_grad_by_freq, self.sparse)

class MetaEmbeddingBag(nn.EmbeddingBag, MetaModule):
    __doc__ = nn.EmbeddingBag.__doc__

    def forward(self, input, offsets=None, per_sample_weights=None, params=None):
        if params is None:
            params = OrderedDict(self.named_parameters())
        return F.embedding_bag(input, params['weight'], offsets,
                               self.max_norm, self.norm_type,
                               self.scale_grad_by_freq, self.mode, self.sparse,
                               per_sample_weights, self.include_last_offset)



import torch
import torch.nn.functional as F
import warnings

from collections import namedtuple
from math import sqrt

__all__ = ['ridge_regression']


_RR_weight_bias = namedtuple('_RR_weight_bias', 'weight bias')

def ridge_regression(embeddings,
                     targets,
                     reg_lambda,
                     num_classes=None,
                     use_woodbury=None,
                     scale=True,
                     bias=True):
    r"""Closed-form solution of a linear function of the embeddings, found 
    using ridge regression.

        W^{*} = argmin_{W} ||XW - Y||^{2} + \lambda ||W||^{2}
              = (X^{T}X + \lambda I)^{-1}X^{T}Y

    Parameters
    ----------
    embeddings : `torch.FloatTensor` instance
        A tensor containing the embeddings of the support points from a single
        task. This tensor has shape `(num_examples, embedding_size)`.

    targets : `torch.LongTensor` or `torch.FloatTensor` instance
        A tensor containing the targets of the support points from a single
        task. If this tensor is a `torch.LongTensor` instance (i.e.
        classification task), then it has shape `(num_examples,)`. If this
        tensor is a `torch.FloatTensor` instance (i.e. regression task), then
        it has shape `(num_examples, output_size)`.

    reg_lambda : float or `torch.FloatTensor` instance
        The regularization constant lambda in the L2-regularized MSE loss. This
        constant can be a tensor.

    num_classes : int (optional)
        Number of classes in the task. If `None`, then `targets` must be a
        `torch.FloatTensor` instance (i.e. regression task).

    use_woodbury : bool (optional)
        If `True`, then the Woodbury formula is used to solve the ridge
        regression. If `None`, then use the Woodbury formula only in cases
        where the number of samples in the support set is smaller than the
        embedding size (standard in few-shot learning).

    scale : bool (default: `True`)
        If `True`, scales the embeddings and targets by the number of
        support points.

    bias : bool (default: `True`)
        If `True`, fit an affine function and return the weights and bias.

    Returns
    -------
    solution : namedtuple instance
        A namedtuple, with fields `weight` and `bias`, containing the weight
        (and bias, if `bias=True`) of the linear function. If `bias=False`,
        then `solution.bias == None`.

    References
    ----------
    .. [1] Bertinetto L., Henriques J. F., Torr P. H.S., Vedaldi A. (2019).
           Meta-learning with differentiable closed-form solvers. In International
           Conference on Learning Representations (https://arxiv.org/abs/1805.08136)
    """
    num_samples, embedding_size = embeddings.size()
    is_regression_task = targets.dtype.is_floating_point
    if num_classes is None:
        if is_regression_task and (targets.ndim != 2):
            raise ValueError('The `targets` tensor has invalid shape `{0}`. '
                             'This tensor must have shape `(num_examples, '
                             'output_size)`.'.format(targets.shape))

        elif not is_regression_task:
            warnings.warn('The number of classes was not given as an input to '
                          '`ridge_regression`. Using the number of classes '
                          'found in `targets`. If you know the number of '
                          'classes (i.e. N in `N-way` classification), then it '
                          'is recommended to provide it to `ridge_regression`.',
                          stacklevel=2)
            num_classes = -1

    if not is_regression_task:
        targets = (F.one_hot(targets, num_classes=num_classes)
                    .to(dtype=embeddings.dtype))

    if use_woodbury is None:
        use_woodbury = (num_samples <= embedding_size + bias)

    if bias:
        ones = embeddings.new_ones((num_samples, 1))
        embeddings = torch.cat([ones, embeddings], dim=1)

    if scale:
        embeddings = embeddings / sqrt(num_samples)
        targets = targets / sqrt(num_samples)

    if use_woodbury:
        eye = torch.eye(num_samples,
                        dtype=embeddings.dtype,
                        device=embeddings.device)
        A = torch.matmul(embeddings, embeddings.t()) + reg_lambda * eye
        solution, _ = torch.solve(targets, A)
        weight_bias = torch.matmul(embeddings.t(), solution)
    else:
        eye = torch.eye(embedding_size + bias,
                        dtype=embeddings.dtype,
                        device=embeddings.device)
        A = torch.matmul(embeddings.t(), embeddings) + reg_lambda * eye
        b = torch.matmul(embeddings.t(), targets)
        weight_bias, _ = torch.solve(b, A)

    if bias:
        return _RR_weight_bias(weight=weight_bias[1:].t(), bias=weight_bias[0])
    else:
        return _RR_weight_bias(weight=weight_bias.t(), bias=None)



import torch.nn as nn
import torch.nn.functional as F

from collections import OrderedDict
from torch.nn.modules.batchnorm import _BatchNorm
from torchmeta.modules.module import MetaModule

class _MetaBatchNorm(_BatchNorm, MetaModule):
    def forward(self, input, params=None):
        self._check_input_dim(input)
        if params is None:
            params = OrderedDict(self.named_parameters())

        # exponential_average_factor is self.momentum set to
        # (when it is available) only so that if gets updated
        # in ONNX graph when this node is exported to ONNX.
        if self.momentum is None:
            exponential_average_factor = 0.0
        else:
            exponential_average_factor = self.momentum

        if self.training and self.track_running_stats:
            if self.num_batches_tracked is not None:
                self.num_batches_tracked += 1
                if self.momentum is None:  # use cumulative moving average
                    exponential_average_factor = 1.0 / float(self.num_batches_tracked)
                else:  # use exponential moving average
                    exponential_average_factor = self.momentum

        weight = params.get('weight', None)
        bias = params.get('bias', None)

        return F.batch_norm(
            input, self.running_mean, self.running_var, weight, bias,
            self.training or not self.track_running_stats,
            exponential_average_factor, self.eps)

class MetaBatchNorm1d(_MetaBatchNorm):
    __doc__ = nn.BatchNorm1d.__doc__

    def _check_input_dim(self, input):
        if input.dim() != 2 and input.dim() != 3:
            raise ValueError('expected 2D or 3D input (got {}D input)'
                             .format(input.dim()))

class MetaBatchNorm2d(_MetaBatchNorm):
    __doc__ = nn.BatchNorm2d.__doc__

    def _check_input_dim(self, input):
        if input.dim() != 4:
            raise ValueError('expected 4D input (got {}D input)'
                             .format(input.dim()))

class MetaBatchNorm3d(_MetaBatchNorm):
    __doc__ = nn.BatchNorm3d.__doc__

    def _check_input_dim(self, input):
        if input.dim() != 5:
            raise ValueError('expected 5D input (got {}D input)'
                             .format(input.dim()))



import torch.nn as nn
import torch.nn.functional as F

from collections import OrderedDict
from torchmeta.modules.module import MetaModule

class MetaLinear(nn.Linear, MetaModule):
    __doc__ = nn.Linear.__doc__

    def forward(self, input, params=None):
        if params is None:
            params = OrderedDict(self.named_parameters())
        bias = params.get('bias', None)
        return F.linear(input, params['weight'], bias)

class MetaBilinear(nn.Bilinear, MetaModule):
    __doc__ = nn.Bilinear.__doc__

    def forward(self, input1, input2, params=None):
        if params is None:
            params = OrderedDict(self.named_parameters())
        bias = params.get('bias', None)
        return F.bilinear(input1, input2, params['weight'], bias)



import numpy as np
from PIL import Image
import os
import io
import json
import glob
import h5py

from torchmeta.utils.data import Dataset, ClassDataset, CombinationMetaDataset
# QKFIX: See torchmeta.datasets.utils for more informations
from torchmeta.datasets.utils import download_file_from_google_drive
from torchmeta.datasets.utils import get_asset


class CUB(CombinationMetaDataset):
    """
    The Caltech-UCSD Birds dataset, introduced in [1]. This dataset is based on
    images from 200 species of birds from the Caltech-UCSD Birds dataset [2].

    Parameters
    ----------
    root : string
        Root directory where the dataset folder `cub` exists.

    num_classes_per_task : int
        Number of classes per tasks. This corresponds to "N" in "N-way" 
        classification.

    meta_train : bool (default: `False`)
        Use the meta-train split of the dataset. If set to `True`, then the
        arguments `meta_val` and `meta_test` must be set to `False`. Exactly one 
        of these three arguments must be set to `True`.

    meta_val : bool (default: `False`)
        Use the meta-validation split of the dataset. If set to `True`, then the 
        arguments `meta_train` and `meta_test` must be set to `False`. Exactly one 
        of these three arguments must be set to `True`.

    meta_test : bool (default: `False`)
        Use the meta-test split of the dataset. If set to `True`, then the 
        arguments `meta_train` and `meta_val` must be set to `False`. Exactly one 
        of these three arguments must be set to `True`.

    meta_split : string in {'train', 'val', 'test'}, optional
        Name of the split to use. This overrides the arguments `meta_train`, 
        `meta_val` and `meta_test` if all three are set to `False`.

    transform : callable, optional
        A function/transform that takes a `PIL` image, and returns a transformed 
        version. See also `torchvision.transforms`.

    target_transform : callable, optional
        A function/transform that takes a target, and returns a transformed 
        version. See also `torchvision.transforms`.

    dataset_transform : callable, optional
        A function/transform that takes a dataset (ie. a task), and returns a 
        transformed version of it. E.g. `torchmeta.transforms.ClassSplitter()`.

    class_augmentations : list of callable, optional
        A list of functions that augment the dataset with new classes. These classes 
        are transformations of existing classes. E.g.
        `torchmeta.transforms.HorizontalFlip()`.

    download : bool (default: `False`)
        If `True`, downloads the pickle files and processes the dataset in the root 
        directory (under the `cub` folder). If the dataset is already 
        available, this does not download/process the dataset again.

    Notes
    -----
    The dataset is downloaded from [2]. The dataset contains images from 200
    classes. The meta train/validation/test splits are over 100/50/50 classes.
    The splits are taken from [3] ([code](https://github.com/wyharveychen/CloserLookFewShot)
    for reproducibility).

    References
    ----------
    .. [1] Hilliard, N., Phillips, L., Howland, S., Yankov, A., Corley, C. D.,
           Hodas, N. O. (2018). Few-Shot Learning with Metric-Agnostic Conditional
           Embeddings. (https://arxiv.org/abs/1802.04376)
    .. [2] Wah, C., Branson, S., Welinder, P., Perona, P., Belongie, S. (2011).
           The Caltech-UCSD Birds-200-2011 Dataset
           (http://www.vision.caltech.edu/visipedia/CUB-200-2011.html)
    .. [3] Chen, W., Liu, Y. and Kira, Z. and Wang, Y. and  Huang, J. (2019).
           A Closer Look at Few-shot Classification. International Conference on
           Learning Representations (https://openreview.net/forum?id=HkxLXnAcFQ)

    """
    def __init__(self, root, num_classes_per_task=None, meta_train=False,
                 meta_val=False, meta_test=False, meta_split=None,
                 transform=None, target_transform=None, dataset_transform=None,
                 class_augmentations=None, download=False):
        dataset = CUBClassDataset(root, meta_train=meta_train, meta_val=meta_val,
            meta_test=meta_test, meta_split=meta_split, transform=transform,
            class_augmentations=class_augmentations, download=download)
        super(CUB, self).__init__(dataset, num_classes_per_task,
            target_transform=target_transform, dataset_transform=dataset_transform)


class CUBClassDataset(ClassDataset):
    folder = 'cub'
    # # Google Drive ID from http://www.vision.caltech.edu/visipedia-data/CUB-200-2011/CUB_200_2011.tgz
    gdrive_id = '1hbzc_P1FuxMkcabkgn9ZKinBwW683j45'
    tgz_filename = 'CUB_200_2011.tgz'
    tgz_md5 = '97eceeb196236b17998738112f37df78'
    image_folder = 'CUB_200_2011/images'

    filename = '{0}_data.hdf5'
    filename_labels = '{0}_labels.json'

    def __init__(self, root, meta_train=False, meta_val=False, meta_test=False,
                 meta_split=None, transform=None, class_augmentations=None,
                 download=False):
        super(CUBClassDataset, self).__init__(meta_train=meta_train,
            meta_val=meta_val, meta_test=meta_test, meta_split=meta_split,
            class_augmentations=class_augmentations)

        self.root = os.path.join(os.path.expanduser(root), self.folder)
        self.transform = transform

        self.split_filename = os.path.join(self.root,
            self.filename.format(self.meta_split))
        self.split_filename_labels = os.path.join(self.root,
            self.filename_labels.format(self.meta_split))

        self._data_file = None
        self._data = None
        self._labels = None

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError('CUB integrity check failed')
        self._num_classes = len(self.labels)

    def __getitem__(self, index):
        label = self.labels[index % self.num_classes]
        data = self.data[label]
        transform = self.get_transform(index, self.transform)
        target_transform = self.get_target_transform(index)

        return CUBDataset(index, data, label, transform=transform,
                          target_transform=target_transform)

    @property
    def num_classes(self):
        return self._num_classes

    @property
    def data(self):
        if self._data is None:
            self._data_file = h5py.File(self.split_filename, 'r')
            self._data = self._data_file['datasets']
        return self._data
    
    @property
    def labels(self):
        if self._labels is None:
            with open(self.split_filename_labels, 'r') as f:
                self._labels = json.load(f)
        return self._labels

    def _check_integrity(self):
        return (os.path.isfile(self.split_filename)
            and os.path.isfile(self.split_filename_labels))

    def close(self):
        if self._data_file is not None:
            self._data_file.close()
            self._data_file = None
            self._data = None

    def download(self):
        import tarfile
        import shutil
        import glob
        from tqdm import tqdm

        if self._check_integrity():
            return

        download_file_from_google_drive(self.gdrive_id, self.root,
            self.tgz_filename, md5=self.tgz_md5)

        tgz_filename = os.path.join(self.root, self.tgz_filename)
        with tarfile.open(tgz_filename, 'r') as f:
            f.extractall(self.root)
        image_folder = os.path.join(self.root, self.image_folder)

        for split in ['train', 'val', 'test']:
            filename = os.path.join(self.root, self.filename.format(split))
            if os.path.isfile(filename):
                continue

            labels = get_asset(self.folder, '{0}.json'.format(split))
            labels_filename = os.path.join(self.root, self.filename_labels.format(split))
            with open(labels_filename, 'w') as f:
                json.dump(labels, f)

            with h5py.File(filename, 'w') as f:
                group = f.create_group('datasets')
                dtype = h5py.special_dtype(vlen=np.uint8)
                for i, label in enumerate(tqdm(labels, desc=filename)):
                    images = glob.glob(os.path.join(image_folder, label, '*.jpg'))
                    images.sort()
                    dataset = group.create_dataset(label, (len(images),), dtype=dtype)
                    for i, image in enumerate(images):
                        with open(image, 'rb') as f:
                            array = bytearray(f.read())
                            dataset[i] = np.asarray(array, dtype=np.uint8)

        tar_folder, _ = os.path.splitext(tgz_filename)
        if os.path.isdir(tar_folder):
            shutil.rmtree(tar_folder)

        attributes_filename = os.path.join(self.root, 'attributes.txt')
        if os.path.isfile(attributes_filename):
            os.remove(attributes_filename)


class CUBDataset(Dataset):
    def __init__(self, index, data, label,
                 transform=None, target_transform=None):
        super(CUBDataset, self).__init__(index, transform=transform,
                                         target_transform=target_transform)
        self.data = data
        self.label = label

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        image = Image.open(io.BytesIO(self.data[index])).convert('RGB')
        target = self.label

        if self.transform is not None:
            image = self.transform(image)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return (image, target)



from torchmeta.modules.activation import MetaMultiheadAttention
from torchmeta.modules.batchnorm import MetaBatchNorm1d, MetaBatchNorm2d, MetaBatchNorm3d
from torchmeta.modules.container import MetaSequential
from torchmeta.modules.conv import MetaConv1d, MetaConv2d, MetaConv3d
from torchmeta.modules.linear import MetaLinear, MetaBilinear
from torchmeta.modules.module import MetaModule
from torchmeta.modules.normalization import MetaLayerNorm
from torchmeta.modules.parallel import DataParallel
from torchmeta.modules.sparse import MetaEmbedding, MetaEmbeddingBag

__all__ = [
    'MetaMultiheadAttention',
    'MetaBatchNorm1d', 'MetaBatchNorm2d', 'MetaBatchNorm3d',
    'MetaSequential',
    'MetaConv1d', 'MetaConv2d', 'MetaConv3d',
    'MetaLinear', 'MetaBilinear',
    'MetaModule',
    'MetaLayerNorm',
    'DataParallel',
    'MetaEmbedding', 'MetaEmbeddingBag',
]


from torchvision.transforms import Compose
from torchmeta.utils.data.task import Task

def apply_wrapper(wrapper, task_or_dataset=None):
    if task_or_dataset is None:
        return wrapper

    from torchmeta.utils.data import MetaDataset
    if isinstance(task_or_dataset, Task):
        return wrapper(task_or_dataset)
    elif isinstance(task_or_dataset, MetaDataset):
        if task_or_dataset.dataset_transform is None:
            dataset_transform = wrapper
        else:
            dataset_transform = Compose([
                task_or_dataset.dataset_transform, wrapper])
        task_or_dataset.dataset_transform = dataset_transform
        return task_or_dataset
    else:
        raise NotImplementedError()

def wrap_transform(transform, fn, transform_type=None):
    if (transform_type is None) or isinstance(transform, transform_type):
        return fn(transform)
    elif isinstance(transform, Compose):
        return Compose([wrap_transform(subtransform, fn, transform_type)
            for subtransform in transform.transforms])
    else:
        return transform



import numpy as np
from PIL import Image
import h5py
import json
import os
import io
import pickle

from torchmeta.utils.data import Dataset, ClassDataset, CombinationMetaDataset
# QKFIX: See torchmeta.datasets.utils for more informations
from torchmeta.datasets.utils import download_file_from_google_drive


class TieredImagenet(CombinationMetaDataset):
    """
    The Tiered-Imagenet dataset, introduced in [1]. This dataset contains images 
    of 608 different classes from the ILSVRC-12 dataset (Imagenet challenge).

    Parameters
    ----------
    root : string
        Root directory where the dataset folder `tieredimagenet` exists.

    num_classes_per_task : int
        Number of classes per tasks. This corresponds to "N" in "N-way" 
        classification.

    meta_train : bool (default: `False`)
        Use the meta-train split of the dataset. If set to `True`, then the
        arguments `meta_val` and `meta_test` must be set to `False`. Exactly one 
        of these three arguments must be set to `True`.

    meta_val : bool (default: `False`)
        Use the meta-validation split of the dataset. If set to `True`, then the 
        arguments `meta_train` and `meta_test` must be set to `False`. Exactly one 
        of these three arguments must be set to `True`.

    meta_test : bool (default: `False`)
        Use the meta-test split of the dataset. If set to `True`, then the 
        arguments `meta_train` and `meta_val` must be set to `False`. Exactly one 
        of these three arguments must be set to `True`.

    meta_split : string in {'train', 'val', 'test'}, optional
        Name of the split to use. This overrides the arguments `meta_train`, 
        `meta_val` and `meta_test` if all three are set to `False`.

    transform : callable, optional
        A function/transform that takes a `PIL` image, and returns a transformed 
        version. See also `torchvision.transforms`.

    target_transform : callable, optional
        A function/transform that takes a target, and returns a transformed 
        version. See also `torchvision.transforms`.

    dataset_transform : callable, optional
        A function/transform that takes a dataset (ie. a task), and returns a 
        transformed version of it. E.g. `torchmeta.transforms.ClassSplitter()`.

    class_augmentations : list of callable, optional
        A list of functions that augment the dataset with new classes. These classes 
        are transformations of existing classes. E.g.
        `torchmeta.transforms.HorizontalFlip()`.

    download : bool (default: `False`)
        If `True`, downloads the pickle files and processes the dataset in the root 
        directory (under the `tieredimagenet` folder). If the dataset is already 
        available, this does not download/process the dataset again.

    Notes
    -----
    The dataset is downloaded from [this repository]
    (https://github.com/renmengye/few-shot-ssl-public/). The dataset contains 
    images from 34 categories. The meta train/validation/test splits are over 
    20/6/8 categories. Each category contains between 10 and 30 classes. The 
    splits over categories (instead of over classes) ensures that all the training 
    classes are sufficiently distinct from the test classes (unlike Mini-Imagenet).

    References
    ----------
    .. [1] Ren, M., Triantafillou, E., Ravi, S., Snell, J., Swersky, K., 
           Tenenbaum, J.B., Larochelle, H. and Zemel, R.S. (2018). Meta-learning 
           for semi-supervised few-shot classification. International Conference 
           on Learning Representations. (https://arxiv.org/abs/1803.00676)
    """
    def __init__(self, root, num_classes_per_task=None, meta_train=False,
                 meta_val=False, meta_test=False, meta_split=None,
                 transform=None, target_transform=None, dataset_transform=None,
                 class_augmentations=None, download=False):
        dataset = TieredImagenetClassDataset(root, meta_train=meta_train,
            meta_val=meta_val, meta_test=meta_test, meta_split=meta_split,
            transform=transform, class_augmentations=class_augmentations,
            download=download)
        super(TieredImagenet, self).__init__(dataset, num_classes_per_task,
            target_transform=target_transform, dataset_transform=dataset_transform)


class TieredImagenetClassDataset(ClassDataset):
    folder = 'tieredimagenet'
    # Google Drive ID from https://github.com/renmengye/few-shot-ssl-public
    gdrive_id = '1g1aIDy2Ar_MViF2gDXFYDBTR-HYecV07'
    tar_filename = 'tiered-imagenet.tar'
    tar_md5 = 'e07e811b9f29362d159a9edd0d838c62'
    tar_folder = 'tiered-imagenet'

    filename = '{0}_data.hdf5'
    filename_labels = '{0}_labels.json'

    def __init__(self, root, meta_train=False, meta_val=False, meta_test=False,
                 meta_split=None, transform=None, class_augmentations=None,
                 download=False):
        super(TieredImagenetClassDataset, self).__init__(meta_train=meta_train,
            meta_val=meta_val, meta_test=meta_test, meta_split=meta_split,
            class_augmentations=class_augmentations)

        self.root = os.path.join(os.path.expanduser(root), self.folder)
        self.transform = transform

        self._data_file = None
        self._data = None
        self._labels = None

        self.split_filename = os.path.join(self.root,
            self.filename.format(self.meta_split))
        self.split_filename_labels = os.path.join(self.root,
            self.filename_labels.format(self.meta_split))

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError('TieredImagenet integrity check failed')
        self._num_classes = len(self.labels)

    @property
    def data(self):
        if self._data is None:
            self._data_file = h5py.File(self.split_filename, 'r')
            self._data = self._data_file['datasets']
        return self._data

    @property
    def labels(self):
        if self._labels is None:
            with open(self.split_filename_labels, 'r') as f:
                self._labels = json.load(f)
        return self._labels

    def __getitem__(self, index):
        specific_class_name = self.labels[index % self.num_classes]
        data = self.data[specific_class_name]
        general_class_name = data.attrs['label_general']
        transform = self.get_transform(index, self.transform)
        target_transform = self.get_target_transform(index)

        return TieredImagenetDataset(index, data,
            general_class_name, specific_class_name,
            transform=transform, target_transform=target_transform)

    @property
    def num_classes(self):
        return self._num_classes

    def close(self):
        if self._data_file is not None:
            self._data_file.close()
            self._data_file = None
            self._data = None

    def _check_integrity(self):
        return (os.path.isfile(self.split_filename)
            and os.path.isfile(self.split_filename_labels))

    def download(self):
        import tarfile
        import shutil
        from tqdm import tqdm

        if self._check_integrity():
            return

        download_file_from_google_drive(self.gdrive_id, self.root,
            self.tar_filename, md5=self.tar_md5)

        filename = os.path.join(self.root, self.tar_filename)
        with tarfile.open(filename, 'r') as f:
            f.extractall(self.root)
        tar_folder = os.path.join(self.root, self.tar_folder)

        for split in ['train', 'val', 'test']:
            filename = os.path.join(self.root, self.filename.format(split))
            if os.path.isfile(filename):
                continue

            images_filename = os.path.join(tar_folder, '{0}_images_png.pkl'.format(split))
            if not os.path.isfile(images_filename):
                raise IOError(images_filename)
            with open(images_filename, 'rb') as f:
                images = pickle.load(f, encoding='bytes')

            labels_filename = os.path.join(tar_folder, '{0}_labels.pkl'.format(split))
            if not os.path.isfile(labels_filename):
                raise IOError()
            with open(labels_filename, 'rb') as f:
                labels = pickle.load(f, encoding='latin1')

            labels_str = labels['label_specific_str']
            general_labels_str = labels['label_general_str']
            general_labels = labels['label_general']
            with open(os.path.join(self.root, self.filename_labels.format(split)), 'w') as f:
                json.dump(labels_str, f)

            with h5py.File(filename, 'w') as f:
                group = f.create_group('datasets')
                dtype = h5py.special_dtype(vlen=np.uint8)
                for i, label in enumerate(tqdm(labels_str, desc=filename)):
                    indices, = np.where(labels['label_specific'] == i)
                    dataset = group.create_dataset(label, (len(indices),), dtype=dtype)
                    general_idx = general_labels[indices[0]]
                    dataset.attrs['label_general'] = (general_labels_str[general_idx]
                        if general_idx < len(general_labels_str) else '')
                    dataset.attrs['label_specific'] = label
                    for j, k in enumerate(indices):
                        dataset[j] = np.squeeze(images[k])

        if os.path.isdir(tar_folder):
            shutil.rmtree(tar_folder)


class TieredImagenetDataset(Dataset):
    def __init__(self, index, data, general_class_name, specific_class_name,
                 transform=None, target_transform=None):
        super(TieredImagenetDataset, self).__init__(index, transform=transform,
                                                    target_transform=target_transform)
        self.data = data
        self.general_class_name = general_class_name
        self.specific_class_name = specific_class_name

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        image = Image.open(io.BytesIO(self.data[index]))
        target = (self.general_class_name, self.specific_class_name)

        if self.transform is not None:
            image = self.transform(image)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return (image, target)



import pytest

import numpy as np
import torch

from torchmeta.utils.matching import (pairwise_cosine_similarity,
    matching_log_probas, matching_probas, matching_loss)


def test_pairwise_similarity():
    eps = 1e-8
    # Numpy
    embeddings1_np = np.random.randn(2, 3, 7).astype(np.float32)
    embeddings2_np = np.random.randn(2, 5, 7).astype(np.float32)

    # PyTorch
    embeddings1_th = torch.as_tensor(embeddings1_np)
    embeddings2_th = torch.as_tensor(embeddings2_np)

    similarities_th = pairwise_cosine_similarity(embeddings1_th, embeddings2_th,
                                                 eps=eps)

    assert similarities_th.shape == (2, 3, 5)

    norm1_np = np.sqrt(np.sum(embeddings1_np ** 2, axis=2))
    norm2_np = np.sqrt(np.sum(embeddings2_np ** 2, axis=2))
    similarities_np = np.zeros((2, 3, 5), dtype=np.float32)
    for i in range(2):
        for j in range(3):
            for k in range(5):
                for l in range(7):
                    similarities_np[i, j, k] += embeddings1_np[i, j, l] * embeddings2_np[i, k, l]
                similarities_np[i, j, k] /= max(norm1_np[i, j] * norm2_np[i, k], eps)
    np.testing.assert_allclose(similarities_th.numpy(), similarities_np, atol=1e-5)


def test_pairwise_similarity_zero():
    eps = 1e-8
    # Numpy
    embeddings1_np = np.random.randn(2, 3, 7).astype(np.float32)
    embeddings2_np = np.random.randn(2, 5, 7).astype(np.float32)

    embeddings1_np[0, 1] = 0.  # Zero out one embedding

    # PyTorch
    embeddings1_th = torch.as_tensor(embeddings1_np)
    embeddings2_th = torch.as_tensor(embeddings2_np)

    similarities_th = pairwise_cosine_similarity(embeddings1_th, embeddings2_th,
                                                 eps=eps)

    assert similarities_th.shape == (2, 3, 5)

    norm1_np = np.sqrt(np.sum(embeddings1_np ** 2, axis=2))
    norm2_np = np.sqrt(np.sum(embeddings2_np ** 2, axis=2))
    similarities_np = np.zeros((2, 3, 5), dtype=np.float32)
    for i in range(2):
        for j in range(3):
            for k in range(5):
                for l in range(7):
                    similarities_np[i, j, k] += embeddings1_np[i, j, l] * embeddings2_np[i, k, l]
                similarities_np[i, j, k] /= max(norm1_np[i, j] * norm2_np[i, k], eps)
    np.testing.assert_allclose(similarities_th.numpy(), similarities_np, atol=1e-5)


def test_matching_log_probas():
    num_classes = 11
    eps = 1e-8
    # Numpy
    embeddings_np = np.random.randn(2, 3, 7).astype(np.float32)
    test_embeddings_np = np.random.randn(2, 5, 7).astype(np.float32)
    targets_np = np.random.randint(num_classes, size=(2, 3))

    # PyTorch
    embeddings_th = torch.as_tensor(embeddings_np)
    test_embeddings_th = torch.as_tensor(test_embeddings_np)
    targets_th = torch.as_tensor(targets_np)

    log_probas_th = matching_log_probas(embeddings_th,
                                        targets_th,
                                        test_embeddings_th,
                                        num_classes,
                                        eps=eps)

    assert log_probas_th.shape == (2, 11, 5)
    np.testing.assert_array_less(log_probas_th.numpy(), 0.)


def test_matching_probas():
    num_classes = 11
    eps = 1e-8
    # Numpy
    embeddings_np = np.random.randn(2, 3, 7).astype(np.float32)
    test_embeddings_np = np.random.randn(2, 5, 7).astype(np.float32)
    targets_np = np.random.randint(num_classes, size=(2, 3))

    # PyTorch
    embeddings_th = torch.as_tensor(embeddings_np)
    test_embeddings_th = torch.as_tensor(test_embeddings_np)
    targets_th = torch.as_tensor(targets_np)

    probas_th = matching_probas(embeddings_th,
                                targets_th,
                                test_embeddings_th,
                                num_classes,
                                eps=eps)

    assert probas_th.shape == (2, 11, 5)
    assert np.all(probas_th.numpy() >= 0.)
    assert np.all(probas_th.numpy() <= 1.)
    np.testing.assert_allclose(probas_th.sum(1).numpy(), 1., atol=1e-5)

    similarities_th = pairwise_cosine_similarity(embeddings_th,
                                                 test_embeddings_th,
                                                 eps=eps)
    exp_similarities_np = np.exp(similarities_th.numpy())

    probas_np = np.zeros((2, 11, 5), dtype=np.float32)
    for i in range(2):
        for j in range(3):
            l = targets_np[i, j]
            for k in range(5):
                probas_np[i, l, k] += (exp_similarities_np[i, j, k]
                    / np.sum(exp_similarities_np[i, :, k]))

    np.testing.assert_allclose(probas_th.numpy(), probas_np, atol=1e-5)



import pytest

import os
import torch
from torch.utils.data import DataLoader

from torchmeta.datasets import helpers
from torchmeta.toy import Sinusoid
from torchmeta.transforms import ClassSplitter
from torchmeta.utils.data import Task, MetaDataLoader, BatchMetaDataLoader

is_local = (os.getenv('TORCHMETA_DATA_FOLDER') is not None)


def test_meta_dataloader():
    dataset = Sinusoid(10, num_tasks=1000, noise_std=None)
    meta_dataloader = MetaDataLoader(dataset, batch_size=4)
    assert isinstance(meta_dataloader, DataLoader)
    assert len(meta_dataloader) == 250 # 1000 / 4
    
    batch = next(iter(meta_dataloader))
    assert isinstance(batch, list)
    assert len(batch) == 4

    task = batch[0]
    assert isinstance(task, Task)
    assert len(task) == 10


def test_meta_dataloader_task_loader():
    dataset = Sinusoid(10, num_tasks=1000, noise_std=None)
    meta_dataloader = MetaDataLoader(dataset, batch_size=4)
    batch = next(iter(meta_dataloader))

    dataloader = DataLoader(batch[0], batch_size=5)
    inputs, targets = next(iter(dataloader))

    assert len(dataloader) == 2 # 10 / 5
    # PyTorch dataloaders convert numpy array to tensors
    assert isinstance(inputs, torch.Tensor)
    assert isinstance(targets, torch.Tensor)
    assert inputs.shape == (5, 1)
    assert targets.shape == (5, 1)


def test_batch_meta_dataloader():
    dataset = Sinusoid(10, num_tasks=1000, noise_std=None)
    meta_dataloader = BatchMetaDataLoader(dataset, batch_size=4)
    assert isinstance(meta_dataloader, DataLoader)
    assert len(meta_dataloader) == 250 # 1000 / 4

    inputs, targets = next(iter(meta_dataloader))
    assert isinstance(inputs, torch.Tensor)
    assert isinstance(targets, torch.Tensor)
    assert inputs.shape == (4, 10, 1)
    assert targets.shape == (4, 10, 1)


def test_batch_meta_dataloader_splitter():
    dataset = Sinusoid(20, num_tasks=1000, noise_std=None)
    dataset = ClassSplitter(dataset, num_train_per_class=5,
        num_test_per_class=15)
    meta_dataloader = BatchMetaDataLoader(dataset, batch_size=4)

    batch = next(iter(meta_dataloader))
    assert isinstance(batch, dict)
    assert 'train' in batch
    assert 'test' in batch

    train_inputs, train_targets = batch['train']
    test_inputs, test_targets = batch['test']

    assert isinstance(train_inputs, torch.Tensor)
    assert isinstance(train_targets, torch.Tensor)
    assert train_inputs.shape == (4, 5, 1)
    assert train_targets.shape == (4, 5, 1)

    assert isinstance(test_inputs, torch.Tensor)
    assert isinstance(test_targets, torch.Tensor)
    assert test_inputs.shape == (4, 15, 1)
    assert test_targets.shape == (4, 15, 1)


@pytest.mark.skipif(not is_local, reason='Requires datasets downloaded locally')
@pytest.mark.parametrize('name', helpers.__all__)
@pytest.mark.parametrize('shots', [1, 5])
@pytest.mark.parametrize('split', ['train', 'val', 'test'])
def test_datasets_helpers_dataloader(name, shots, split):
    function = getattr(helpers, name)
    folder = os.getenv('TORCHMETA_DATA_FOLDER')
    download = bool(os.getenv('TORCHMETA_DOWNLOAD', False))

    dataset = function(folder,
                       ways=5,
                       shots=shots,
                       test_shots=15,
                       meta_split=split,
                       download=download)

    meta_dataloader = BatchMetaDataLoader(dataset, batch_size=4)

    batch = next(iter(meta_dataloader))
    assert isinstance(batch, dict)
    assert 'train' in batch
    assert 'test' in batch

    train_inputs, train_targets = batch['train']
    test_inputs, test_targets = batch['test']

    assert isinstance(train_inputs, torch.Tensor)
    assert isinstance(train_targets, torch.Tensor)
    assert train_inputs.ndim == 5
    assert train_inputs.shape[:2] == (4, 5 * shots)
    assert train_targets.ndim == 2
    assert train_targets.shape[:2] == (4, 5 * shots)

    assert isinstance(test_inputs, torch.Tensor)
    assert isinstance(test_targets, torch.Tensor)
    assert test_inputs.ndim == 5
    assert test_inputs.shape[:2] == (4, 5 * 15) # test_shots
    assert test_targets.ndim == 2
    assert test_targets.shape[:2] == (4, 5 * 15)


@pytest.mark.skipif(not is_local, reason='Requires datasets downloaded locally')
def test_overflow_length_dataloader():
    folder = os.getenv('TORCHMETA_DATA_FOLDER')
    download = bool(os.getenv('TORCHMETA_DOWNLOAD', False))

    # The number of tasks is C(4112, 20), which exceeds machine precision
    dataset = helpers.omniglot(folder,
                               ways=20,
                               shots=1,
                               test_shots=5,
                               meta_train=True,
                               download=download)

    meta_dataloader = BatchMetaDataLoader(dataset, batch_size=4)

    batch = next(iter(meta_dataloader))
    assert isinstance(batch, dict)
    assert 'train' in batch
    assert 'test' in batch



import os
import json


def get_asset_path(*args):
    basedir = os.path.dirname(__file__)
    return os.path.join(basedir, 'assets', *args)


def get_asset(*args, dtype=None):
    filename = get_asset_path(*args)
    if not os.path.isfile(filename):
        raise IOError('{} not found'.format(filename))

    if dtype is None:
        _, dtype = os.path.splitext(filename)
        dtype = dtype[1:]

    if dtype == 'json':
        with open(filename, 'r') as f:
            data = json.load(f)
    else:
        raise NotImplementedError()
    return data

# QKFIX: The current version of `download_file_from_google_drive` (as of torchvision==0.8.1)
# is inconsistent, and a temporary fix has been added to the bleeding-edge version of
# Torchvision. The temporary fix removes the behaviour of `_quota_exceeded`, whenever the
# quota has exceeded for the file to be downloaded. As a consequence, this means that there
# is currently no protection against exceeded quotas. If you get an integrity error in Torchmeta
# (e.g. "MiniImagenet integrity check failed" for MiniImagenet), then this means that the quota
# has exceeded for this dataset. See also: https://github.com/tristandeleu/pytorch-meta/issues/54
# 
# See also: https://github.com/pytorch/vision/issues/2992
# 
# The following functions are taken from
# https://github.com/pytorch/vision/blob/cd0268cd408d19d91f870e36fdffd031085abe13/torchvision/datasets/utils.py

from torchvision.datasets.utils import _get_confirm_token, _save_response_content

def _quota_exceeded(response: "requests.models.Response"):
    return False
    # See https://github.com/pytorch/vision/issues/2992 for details
    # return "Google Drive - Quota exceeded" in response.text


def download_file_from_google_drive(file_id, root, filename=None, md5=None):
    """Download a Google Drive file from  and place it in root.

    Args:
        file_id (str): id of file to be downloaded
        root (str): Directory to place downloaded file in
        filename (str, optional): Name to save the file under. If None, use the id of the file.
        md5 (str, optional): MD5 checksum of the download. If None, do not check
    """
    # Based on https://stackoverflow.com/questions/38511444/python-download-files-from-google-drive-using-url
    import requests
    url = "https://docs.google.com/uc?export=download"

    root = os.path.expanduser(root)
    if not filename:
        filename = file_id
    fpath = os.path.join(root, filename)

    os.makedirs(root, exist_ok=True)

    if os.path.isfile(fpath) and check_integrity(fpath, md5):
        print('Using downloaded and verified file: ' + fpath)
    else:
        session = requests.Session()

        response = session.get(url, params={'id': file_id}, stream=True)
        token = _get_confirm_token(response)

        if token:
            params = {'id': file_id, 'confirm': token}
            response = session.get(url, params=params, stream=True)

        if _quota_exceeded(response):
            msg = (
                f"The daily quota of the file {filename} is exceeded and it "
                f"can't be downloaded. This is a limitation of Google Drive "
                f"and can only be overcome by trying again later."
            )
            raise RuntimeError(msg)

        _save_response_content(response, fpath)



from torchmeta.transforms.categorical import Categorical, FixedCategory
from torchmeta.transforms.augmentations import Rotation, HorizontalFlip, VerticalFlip
from torchmeta.transforms.splitters import Splitter, ClassSplitter, WeightedClassSplitter
from torchmeta.transforms.target_transforms import TargetTransform, DefaultTargetTransform, SegmentationPairTransform



from torchmeta.utils.data.dataloader import MetaDataLoader, BatchMetaDataLoader
from torchmeta.utils.data.dataset import ClassDataset, MetaDataset, CombinationMetaDataset
from torchmeta.utils.data.sampler import CombinationSequentialSampler, CombinationRandomSampler
from torchmeta.utils.data.task import Dataset, Task, ConcatTask, SubsetTask
from torchmeta.utils.data.wrappers import NonEpisodicWrapper

__all__ = [
    'MetaDataLoader',
    'BatchMetaDataLoader',
    'ClassDataset',
    'MetaDataset',
    'CombinationMetaDataset',
    'CombinationSequentialSampler',
    'CombinationRandomSampler',
    'Dataset',
    'Task',
    'ConcatTask',
    'SubsetTask',
    'NonEpisodicWrapper'
]



import os
import pickle
from PIL import Image
import h5py
import json

from torchmeta.utils.data import Dataset, ClassDataset, CombinationMetaDataset
# QKFIX: See torchmeta.datasets.utils for more informations
from torchmeta.datasets.utils import download_file_from_google_drive


class MiniImagenet(CombinationMetaDataset):
    """
    The Mini-Imagenet dataset, introduced in [1]. This dataset contains images 
    of 100 different classes from the ILSVRC-12 dataset (Imagenet challenge). 
    The meta train/validation/test splits are taken from [2] for reproducibility.

    Parameters
    ----------
    root : string
        Root directory where the dataset folder `miniimagenet` exists.

    num_classes_per_task : int
        Number of classes per tasks. This corresponds to "N" in "N-way" 
        classification.

    meta_train : bool (default: `False`)
        Use the meta-train split of the dataset. If set to `True`, then the
        arguments `meta_val` and `meta_test` must be set to `False`. Exactly one 
        of these three arguments must be set to `True`.

    meta_val : bool (default: `False`)
        Use the meta-validation split of the dataset. If set to `True`, then the 
        arguments `meta_train` and `meta_test` must be set to `False`. Exactly one 
        of these three arguments must be set to `True`.

    meta_test : bool (default: `False`)
        Use the meta-test split of the dataset. If set to `True`, then the 
        arguments `meta_train` and `meta_val` must be set to `False`. Exactly one 
        of these three arguments must be set to `True`.

    meta_split : string in {'train', 'val', 'test'}, optional
        Name of the split to use. This overrides the arguments `meta_train`, 
        `meta_val` and `meta_test` if all three are set to `False`.

    transform : callable, optional
        A function/transform that takes a `PIL` image, and returns a transformed 
        version. See also `torchvision.transforms`.

    target_transform : callable, optional
        A function/transform that takes a target, and returns a transformed 
        version. See also `torchvision.transforms`.

    dataset_transform : callable, optional
        A function/transform that takes a dataset (ie. a task), and returns a 
        transformed version of it. E.g. `torchmeta.transforms.ClassSplitter()`.

    class_augmentations : list of callable, optional
        A list of functions that augment the dataset with new classes. These classes 
        are transformations of existing classes. E.g.
        `torchmeta.transforms.HorizontalFlip()`.

    download : bool (default: `False`)
        If `True`, downloads the pickle files and processes the dataset in the root 
        directory (under the `miniimagenet` folder). If the dataset is already 
        available, this does not download/process the dataset again.

    Notes
    -----
    The dataset is downloaded from [this repository]
    (https://github.com/renmengye/few-shot-ssl-public/). The meta train/
    validation/test splits are over 64/16/20 classes.

    References
    ----------
    .. [1] Vinyals, O., Blundell, C., Lillicrap, T. and Wierstra, D. (2016). 
           Matching Networks for One Shot Learning. In Advances in Neural 
           Information Processing Systems (pp. 3630-3638) (https://arxiv.org/abs/1606.04080)

    .. [2] Ravi, S. and Larochelle, H. (2016). Optimization as a Model for 
           Few-Shot Learning. (https://openreview.net/forum?id=rJY0-Kcll)
    """
    def __init__(self, root, num_classes_per_task=None, meta_train=False,
                 meta_val=False, meta_test=False, meta_split=None,
                 transform=None, target_transform=None, dataset_transform=None,
                 class_augmentations=None, download=False):
        dataset = MiniImagenetClassDataset(root, meta_train=meta_train,
            meta_val=meta_val, meta_test=meta_test, meta_split=meta_split,
            transform=transform, class_augmentations=class_augmentations,
            download=download)
        super(MiniImagenet, self).__init__(dataset, num_classes_per_task,
            target_transform=target_transform, dataset_transform=dataset_transform)


class MiniImagenetClassDataset(ClassDataset):
    folder = 'miniimagenet'
    # Google Drive ID from https://github.com/renmengye/few-shot-ssl-public
    gdrive_id = '16V_ZlkW4SsnNDtnGmaBRq2OoPmUOc5mY'
    gz_filename = 'mini-imagenet.tar.gz'
    gz_md5 = 'b38f1eb4251fb9459ecc8e7febf9b2eb'
    pkl_filename = 'mini-imagenet-cache-{0}.pkl'

    filename = '{0}_data.hdf5'
    filename_labels = '{0}_labels.json'

    def __init__(self, root, meta_train=False, meta_val=False, meta_test=False,
                 meta_split=None, transform=None, class_augmentations=None,
                 download=False):
        super(MiniImagenetClassDataset, self).__init__(meta_train=meta_train,
            meta_val=meta_val, meta_test=meta_test, meta_split=meta_split,
            class_augmentations=class_augmentations)
        
        self.root = os.path.join(os.path.expanduser(root), self.folder)
        self.transform = transform

        self.split_filename = os.path.join(self.root,
            self.filename.format(self.meta_split))
        self.split_filename_labels = os.path.join(self.root,
            self.filename_labels.format(self.meta_split))

        self._data = None
        self._labels = None

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError('MiniImagenet integrity check failed')
        self._num_classes = len(self.labels)

    def __getitem__(self, index):
        class_name = self.labels[index % self.num_classes]
        data = self.data[class_name]
        transform = self.get_transform(index, self.transform)
        target_transform = self.get_target_transform(index)

        return MiniImagenetDataset(index, data, class_name,
            transform=transform, target_transform=target_transform)

    @property
    def num_classes(self):
        return self._num_classes

    @property
    def data(self):
        if self._data is None:
            self._data_file = h5py.File(self.split_filename, 'r')
            self._data = self._data_file['datasets']
        return self._data

    @property
    def labels(self):
        if self._labels is None:
            with open(self.split_filename_labels, 'r') as f:
                self._labels = json.load(f)
        return self._labels

    def _check_integrity(self):
        return (os.path.isfile(self.split_filename)
            and os.path.isfile(self.split_filename_labels))

    def close(self):
        if self._data_file is not None:
            self._data_file.close()
            self._data_file = None
            self._data = None

    def download(self):
        import tarfile

        if self._check_integrity():
            return

        download_file_from_google_drive(self.gdrive_id, self.root,
            self.gz_filename, md5=self.gz_md5)

        filename = os.path.join(self.root, self.gz_filename)
        with tarfile.open(filename, 'r') as f:
            f.extractall(self.root)

        for split in ['train', 'val', 'test']:
            filename = os.path.join(self.root, self.filename.format(split))
            if os.path.isfile(filename):
                continue

            pkl_filename = os.path.join(self.root, self.pkl_filename.format(split))
            if not os.path.isfile(pkl_filename):
                raise IOError()
            with open(pkl_filename, 'rb') as f:
                data = pickle.load(f)
                images, classes = data['image_data'], data['class_dict']

            with h5py.File(filename, 'w') as f:
                group = f.create_group('datasets')
                for name, indices in classes.items():
                    group.create_dataset(name, data=images[indices])

            labels_filename = os.path.join(self.root, self.filename_labels.format(split))
            with open(labels_filename, 'w') as f:
                labels = sorted(list(classes.keys()))
                json.dump(labels, f)

            if os.path.isfile(pkl_filename):
                os.remove(pkl_filename)


class MiniImagenetDataset(Dataset):
    def __init__(self, index, data, class_name,
                 transform=None, target_transform=None):
        super(MiniImagenetDataset, self).__init__(index, transform=transform,
                                                  target_transform=target_transform)
        self.data = data
        self.class_name = class_name

    def __len__(self):
        return self.data.shape[0]

    def __getitem__(self, index):
        image = Image.fromarray(self.data[index])
        target = self.class_name

        if self.transform is not None:
            image = self.transform(image)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return (image, target)



import numpy as np
import os
import json
import h5py
from PIL import Image

from torchvision.datasets.utils import check_integrity, download_url
from torchmeta.utils.data import Dataset, ClassDataset


class CIFAR100ClassDataset(ClassDataset):
    folder = 'cifar100'
    subfolder = None
    download_url = 'https://www.cs.toronto.edu/~kriz/cifar-100-python.tar.gz'
    gz_folder = 'cifar-100-python'
    gz_md5 = 'eb9058c3a382ffc7106e4002c42a8d85'
    files_md5 = {
        'train': '16019d7e3df5f24257cddd939b257f8d',
        'test': 'f0ef6b0ae62326f3e7ffdfab6717acfc',
        'meta': '7973b15100ade9c7d40fb424638fde48'
    }

    filename = 'data.hdf5'
    filename_labels = '{0}_labels.json'
    filename_fine_names = 'fine_names.json'

    def __init__(self, root, meta_train=False, meta_val=False, meta_test=False,
                 meta_split=None, transform=None, class_augmentations=None,
                 download=False):
        super(CIFAR100ClassDataset, self).__init__(meta_train=meta_train,
            meta_val=meta_val, meta_test=meta_test, meta_split=meta_split,
            class_augmentations=class_augmentations)

        if self.subfolder is None:
            raise ValueError()

        self.root = os.path.join(os.path.expanduser(root), self.folder)
        self.transform = transform

        self.split_filename_labels = os.path.join(self.root, self.subfolder,
            self.filename_labels.format(self.meta_split))
        self._data = None
        self._labels = None

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError('CIFAR100 integrity check failed')
        self._num_classes = len(self.labels)

    def __getitem__(self, index):
        coarse_label_name, fine_label_name = self.labels[index % self.num_classes]
        data = self.data['{0}/{1}'.format(coarse_label_name, fine_label_name)]
        transform = self.get_transform(index, self.transform)
        target_transform = self.get_target_transform(index)

        return CIFAR100Dataset(index, data, coarse_label_name, fine_label_name,
            transform=transform, target_transform=target_transform)

    @property
    def num_classes(self):
        return self._num_classes

    @property
    def data(self):
        if self._data is None:
            self._data = h5py.File(os.path.join(self.root, self.filename), 'r')
        return self._data

    @property
    def labels(self):
        if self._labels is None:
            with open(self.split_filename_labels, 'r') as f:
                self._labels = json.load(f)
        return self._labels

    def _check_integrity(self):
        return (self._check_integrity_data()
            and os.path.isfile(self.split_filename_labels)
            and os.path.isfile(os.path.join(self.root, self.filename_fine_names)))

    def _check_integrity_data(self):
        return os.path.isfile(os.path.join(self.root, self.filename))

    def close(self):
        if self._data is not None:
            self._data.close()
            self._data = None

    def download(self):
        import tarfile
        import pickle
        import shutil

        if self._check_integrity_data():
            return

        gz_filename = '{0}.tar.gz'.format(self.gz_folder)
        download_url(self.download_url, self.root, filename=gz_filename,
                     md5=self.gz_md5)
        with tarfile.open(os.path.join(self.root, gz_filename), 'r:gz') as tar:
            tar.extractall(path=self.root)

        train_filename = os.path.join(self.root, self.gz_folder, 'train')
        check_integrity(train_filename, self.files_md5['train'])
        with open(train_filename, 'rb') as f:
            data = pickle.load(f, encoding='bytes')
            images = data[b'data']
            fine_labels = data[b'fine_labels']
            coarse_labels = data[b'coarse_labels']

        test_filename = os.path.join(self.root, self.gz_folder, 'test')
        check_integrity(test_filename, self.files_md5['test'])
        with open(test_filename, 'rb') as f:
            data = pickle.load(f, encoding='bytes')
            images = np.concatenate((images, data[b'data']), axis=0)
            fine_labels = np.concatenate((fine_labels, data[b'fine_labels']), axis=0)
            coarse_labels = np.concatenate((coarse_labels, data[b'coarse_labels']), axis=0)

        images = images.reshape((-1, 3, 32, 32))
        images = images.transpose((0, 2, 3, 1))

        meta_filename = os.path.join(self.root, self.gz_folder, 'meta')
        check_integrity(meta_filename, self.files_md5['meta'])
        with open(meta_filename, 'rb') as f:
            data = pickle.load(f, encoding='latin1')
            fine_label_names = data['fine_label_names']
            coarse_label_names = data['coarse_label_names']

        filename = os.path.join(self.root, self.filename)
        fine_names = dict()
        with h5py.File(filename, 'w') as f:
            for i, coarse_name in enumerate(coarse_label_names):
                group = f.create_group(coarse_name)
                fine_indices = np.unique(fine_labels[coarse_labels == i])
                for j in fine_indices:
                    dataset = group.create_dataset(fine_label_names[j],
                        data=images[fine_labels == j])
                fine_names[coarse_name] = [fine_label_names[j] for j in fine_indices]

        filename_fine_names = os.path.join(self.root, self.filename_fine_names)
        with open(filename_fine_names, 'w') as f:
            json.dump(fine_names, f)

        gz_folder = os.path.join(self.root, self.gz_folder)
        if os.path.isdir(gz_folder):
            shutil.rmtree(gz_folder)
        if os.path.isfile('{0}.tar.gz'.format(gz_folder)):
            os.remove('{0}.tar.gz'.format(gz_folder))


class CIFAR100Dataset(Dataset):
    def __init__(self, index, data, coarse_label_name, fine_label_name,
                 transform=None, target_transform=None):
        super(CIFAR100Dataset, self).__init__(index, transform=transform,
                                              target_transform=target_transform)
        self.data = data
        self.coarse_label_name = coarse_label_name
        self.fine_label_name = fine_label_name

    def __len__(self):
        return self.data.shape[0]

    def __getitem__(self, index):
        image = Image.fromarray(self.data[index])
        target = (self.coarse_label_name, self.fine_label_name)

        if self.transform is not None:
            image = self.transform(image)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return (image, target)



import numpy as np

from torchmeta.utils.data import Task, MetaDataset


class Harmonic(MetaDataset):
    """
    Simple regression task, based on the sum of two sine waves, as introduced
    in [1].

    Parameters
    ----------
    num_samples_per_task : int
        Number of examples per task.

    num_tasks : int (default: 5,000)
        Overall number of tasks to sample.

    noise_std : float, optional
        Amount of noise to include in the targets for each task. If `None`, then
        nos noise is included, and the target is the sum of 2 sine functions of
        the input.

    transform : callable, optional
        A function/transform that takes a numpy array of size (1,) and returns a
        transformed version of the input.

    target_transform : callable, optional
        A function/transform that takes a numpy array of size (1,) and returns a
        transformed version of the target.

    dataset_transform : callable, optional
        A function/transform that takes a dataset (ie. a task), and returns a 
        transformed version of it. E.g. `torchmeta.transforms.ClassSplitter()`.

    Notes
    -----
    The tasks are created randomly as the sum of two sinusoid functions, with
    a frequency ratio of 2. The amplitudes vary within [5.0, 7.0], the phases
    within [0, 2 * pi], and the inputs are sampled according to N(mu_x, 1), with
    mu_x varying in [-4.0, 4.0]. Due to the way PyTorch handles datasets, the
    number of tasks to be sampled needs to be fixed ahead of time (with
    `num_tasks`). This will typically be equal to `meta_batch_size * num_batches`.

    References
    ----------
    .. [1] Lacoste A., Oreshkin B., Chung W., Boquet T., Rostamzadeh N.,
           Krueger D. (2018). Uncertainty in Multitask Transfer Learning. In
           Advances in Neural Information Processing Systems (https://arxiv.org/abs/1806.07528)
    """
    def __init__(self, num_samples_per_task, num_tasks=5000,
                 noise_std=None, transform=None, target_transform=None,
                 dataset_transform=None):
        super(Harmonic, self).__init__(meta_split='train',
            target_transform=target_transform, dataset_transform=dataset_transform)
        self.num_samples_per_task = num_samples_per_task
        self.num_tasks = num_tasks
        self.noise_std = noise_std
        self.transform = transform

        self._domain_range = np.array([-4.0, 4.0])
        self._frequency_range = np.array([5.0, 7.0])
        self._phase_range = np.array([0, 2 * np.pi])

        self._domains = None
        self._frequencies = None
        self._phases = None
        self._amplitudes = None

    @property
    def domains(self):
        if self._domains is None:
            self._domains = self.np_random.uniform(self._domain_range[0],
                self._domain_range[1], size=self.num_tasks)
        return self._domains

    @property
    def frequencies(self):
        if self._frequencies is None:
            self._frequencies = self.np_random.uniform(self._frequency_range[0],
                self._frequency_range[1], size=self.num_tasks)
        return self._frequencies

    @property
    def phases(self):
        if self._phases is None:
            self._phases = self.np_random.uniform(self._phase_range[0],
                self._phase_range[1], size=(self.num_tasks, 2))
        return self._phases

    @property
    def amplitudes(self):
        if self._amplitudes is None:
            self._amplitudes = self.np_random.randn(self.num_tasks, 2)
        return self._amplitudes

    def __len__(self):
        return self.num_tasks

    def __getitem__(self, index):
        domain = self.domains[index]
        frequency = self.frequencies[index]
        phases = self.phases[index]
        amplitudes = self.amplitudes[index]

        task = HarmonicTask(index, domain, frequency, phases, amplitudes,
            self.noise_std, self.num_samples_per_task, self.transform,
            self.target_transform, np_random=self.np_random)

        if self.dataset_transform is not None:
            task = self.dataset_transform(task)

        return task


class HarmonicTask(Task):
    def __init__(self, index, domain, frequency, phases, amplitudes,
                 noise_std, num_samples, transform=None,
                 target_transform=None, np_random=None):
        super(HarmonicTask, self).__init__(index, None) # Regression task
        self.domain = domain
        self.frequency = frequency
        self.phases = phases
        self.amplitudes = amplitudes
        self.noise_std = noise_std
        self.num_samples = num_samples

        self.transform = transform
        self.target_transform = target_transform

        if np_random is None:
            np_random = np.random.RandomState(None)

        a_1, a_2 = self.amplitudes
        b_1, b_2 = self.phases

        self._inputs = self.domain + np_random.randn(num_samples, 1)
        self._targets = (a_1 * np.sin(frequency * self._inputs + b_1)
            + a_2 * np.sin(2 * frequency * self._inputs + b_2))
        if (noise_std is not None) and (noise_std > 0.):
            self._targets += noise_std * np_random.randn(num_samples, 1)

    def __len__(self):
        return self.num_samples

    def __getitem__(self, index):
        input, target = self._inputs[index], self._targets[index]

        if self.transform is not None:
            input = self.transform(input)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return (input, target)
